//
//  Font base
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitFontBase_hpp
#define GuitFontBase_hpp
#include <gtypes.hpp>
namespace guit {

class GFontFace {
public:
  GFontFace(GString const& family, GString const& style, GString const& path);
  GString family_, style_, path_;
};


/// Font data base.
/// The font data base is initialized by reading the Guit font cache file (~/Guit/font-cache).
/// This file is automatically created the first time a Guit applicatiin is launched by searching for
/// TrueType (.ttf) font files in the directories specified by GPrefs::fontDirs()
/// To re-generate the font cache (e.g. when new fonts are added), just delete this file.
class GFontBase {
public:
  /// Searches a font face in the font base.
  static GFontFace* findFontFace(GString const& family, GString const& style);

private:
  static void init();

  /// Searches TrueType fonts on the filesystem and creates the font list file.
  /// Called when launching a Guit app if the font list does not already exist.
  /// - searches TrueType font files in the directories specified by GPrefs::fontDirs()
  /// - calls fc-list if GPrefs::fontDirs() is empty
  /// - creates ~/.guit/font-list.txt
  static bool makeFontList(GString const& guitdir);

  /// Reads the font list file to initialize the font subsystem.
  static bool readFontList(GString const& guitdir);

  /// Called by readFontList() to add font faces to the font subsystem.
  static bool addFontFace(GString const& font_definition);
};


class GFontResource {
public:
  ~GFontResource();

  /// Returns the first available font matching these characteristics.
  /// - Font resources are shared: creates a new one only if needed
  /// - families ca be a list, eg: "Arial, Helvetica, sans-serif"
  /// - note that sans-serif (or serif or monospace) should always be present to ensure compatibility
  ///  with various platforms
  static GFontResource* obtainFont(GString const& style, uint8_t size,
                                   GString const& families);

  /// returns the empty font, used by no-arg constructors, etc.
  static GFontResource* emptyFont();

  static void set(GFontResource*&, GFontResource*);

  /// Allocates resources, returns true if resources could be allocated.
  /// does not reallocate resources if already done.
  bool realize();

  uint8_t size() const {return size_;}
  GString const& style() const {return style_;}
  GString const& families() const {return families_;}
  class GNatFont* natfont() const {return natfont_;}

private:
  bool cantcreate_{};
  uint8_t size_{};
  uint16_t refcount_{};
  GFontFace* face_{};
  GString style_, families_;
 class GNatFont* natfont_{};

private:
  friend class GNatContext;
  GFontResource(GFontFace*, GString const& style, uint8_t size, GString const& families);
};

}
#endif
