//
//  Styles & CSS
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Style_hpp
#define Guit_Style_hpp
#include <list>
#include <unordered_map>
#include <gprop.hpp>
#include <gadget.hpp>
namespace guit {

struct GStyleMap : public std::unordered_map<GId,GStyle*> {
  ~GStyleMap();
  GStyle* findStyle(GId selid);
  GStyle& obtainStyle(GId selid);
};

class GPseudoStyle;

/// Collection of props.
/// - Initially, the props are  stored textually
/// - They are transformed into GProp(s) when realize() is called. realize() must NOT be called
///  before GApp is initialized (because fonts and other resources wont be initialized)
/// - the last prop is kept in case of mulipe additions of the same kind of prop (see GProp::role())
/// @see GTheme.
class GStyle {
public:
  enum Specif : GProp::Specif {
    TypeSpecif =           1,
    ClassSpecif =        100,
    IdSpecif =        10'000,
    InlineSpecif = 1'000'000,
    InterSpecif = 10'000'000,
    CondSpecif =  20'000'000,
    AnimSpecif =  30'000'000
  };

  struct PropMap : public std::unordered_map<int, gptr<GProp>> {
    void apply(Gadget* g, GRender& r, GProp::Specif specif) {
      for (auto& p : *this) p.second->apply(g, r, specif);
    }
  };

  ~GStyle();
  GStyle(GId selector_id);

  /// creates props.
  void realize();

  /// adds prop, the new prop replaces the previous equivalent prop if any (see GProp).
  /// @{
  void addProp(GProp* prop);
  GStyle& operator<<(GProp* prop) {addProp(prop); return *this;}
  GStyle& operator<<(GProp& prop) {addProp(&prop); return *this;}
  /// @}

  /// finds prop of this class.
  /// @{
  GProp* findProp(GPropType&) const;
  template <class T> auto* findProp() const {return dynamic_cast<T*>(findProp(T::Type()));}
  /// @}

  /// removes prop, or prop of this class.
  /// @{
  void removeProp(GProp&);
  void removeProp(GPropType&);
  /// @}

  PropMap const& props() const {return props_;}

  void apply(Gadget* g, GRender& r) {
    if (!realized_) realize();
    if (!props_.empty()) props_.apply(g, r, specif_);
    if (!pseudos_.empty()) applyPseudos(g, r);
  }

  void applyPseudos(Gadget* g, GRender& r);

  // calls realize() if style if alraedy realized_, otherwise just adds decl block
  void addStyleDecl(GString const& decl, GProp::Specif, GPseudoStyle* ps);
  GStyle* addStyleDecl2(GString const& decl, GProp::Specif, GPseudoStyle* ps,
                        GString const& desc, char reltype, bool is_subrule);

  /// @internal
  /// @{
  enum RelType {DirectChild = 1};
  void addDecl(GString const& decl);
  bool realized_{}, has_desc_{}, any_desc_{}, direct_desc_{}, has_subrule_{}, is_pseudo_{};
  GId id_{};
  GProp::Specif specif_{};
  GadgetModes modes_{};
  GString decl_;
  PropMap props_;
  GStyleMap* stylemap_{};
  std::forward_list<GPseudoStyle*> pseudos_;
  /// @}
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Style Base.
class GStyleBase {
public:
  /// returns the style base of the application.
  static GStyleBase& instance();

  /// loads a CSS style from a file (see GIn for magic prefix characters).
  bool loadFile(GString const& css_filename);

  /// loads a CSS style from a text buffer.
  bool loadBuffer(GString const& css_buffer);

  /// loads a CSS style from an input stream (stream_name only used for displaying errors).
  bool load(std::istream& in, GString const& stream_name);

  /// finds a style.
  /// @note the style may not be initialized, GStyle::realize() must then be called.
  /// @{
  GStyle* findStyle(GId selector_id);
  GStyle* findStyle(GString const& selector_name);
  GStyle& obtainStyle(GId selector_id);
  GStyle& obtainStyle(GString const& selector_name);
  /// @}

  /// adds a style .
  bool addStyle(GStrings& seltab, GString const& decl_block);

  /// @internal
  /// @{
  void addCoreStyle(GString const& name, GStyle&);
  void error(GString const& msg);
  bool parseStyle(std::istream&);


  //bool parseSelName(GString const& sel, GString& name, 
  //                  GStrings& pseudos, GProp::Specif&);


  bool parseSelName(GString const& sel, GString& name,
                                GProp::Specif& spec, GPseudoStyle*& ps);

  GString streamname_;
  GStyleMap styles_;
  /// @}
};


}
#endif
