//
// GBox: main container gadgets
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <algorithm>
#include <gevent.hpp>
#include <gtrigger.hpp>
#include <gborder.hpp>
#include <gbox.hpp>
#include <gexpr.hpp>
#include <gvalue.hpp>
#include <gtextitem.hpp>
#include <gchoice.hpp>
#include <gmenu.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(Box)
GIMPLEMENT_GADGET(HBox)
GIMPLEMENT_GADGET(Pane)
GIMPLEMENT_GADGET(ButtonBox)

class GBoxViews : public GProp {
public:
  using Value = std::vector<GBox*>;

  GString stringValue() const override {return "<BoxViews>";}

  static GPropType& Type() {
    static auto& t = *new GPropType("BoxViews", GPropType::Auto);
    return t;
  }
  GPropType& type() const override {return Type();}

  Value views_;
};


GBox::~GBox() {
  if (settings_.hasViews) {
    if (GBoxViews* views = findProp<GBoxViews>()) {
      auto& list = views->views_;
      // removes this view from the list of views
      auto it = std::find(list.begin(), list.end(), this);
      if (it != list.end()) {
        list.erase(it);
      }
    }
  }
  removeChildren(Remove(Remove::InContent | Remove::InBorder | Remove::Internals));
}

GBox::GBox(GadgetType& t, GString const& spec) : Gadget(t) {
  if (!spec.empty()) {
    GString title;
    GadgetStyle::initStyle(*this, spec, title);
    if (!title.empty()) add(new GTitle(title));
  }
}

GBox::GBox(GString const& spec)
: GBox(Type(), spec) {}

GChoice* GBox::choiceProp() {return findProp<GChoice>();}
GProp* GBox::valueProp() {return findProp<GChoice>();}

GBox& GBox::clone(GClone const& cm) {   // REVOIR LES clone()  !!!!
  auto* g = new GBox("");      // MANQUE PLEIN DE CLASSES
  g->initClone(*this, cm);              // ET LA SPEC EST PERDUE !!!
  return *g;
}


void GBox::initClone(Gadget& from, GClone const& c) {
  // as the item are cloned they must not be created from styles
  settings_.itemsSet = 1;

  GBox* frombox = from.toBox();
  if (!frombox) {  // should not occuer
    error("clone","internal error: not a GBox");
    return;
  }

  if (c & GClone::Sync) {
    GBoxViews* views{};
    // if from already has views
    if (from.settings_.hasViews) views = from.findProp<GBoxViews>();
    // otherwise adds views to from
    if (!views) {
      views = new GBoxViews;
      views->views_.push_back(frombox);
      from.add(*views);
      from.settings_.hasViews = 1;
    }
    // clones this and its descendants. beware the clone will share the BoxViews
    // prop and have its hasViews mode set to the same value as this.
    // If this->hasViews was true, the clone would thus duplicate its children
    // in this (as this is the expected effect of hasViews!), meaning that the
    // children of this would be duplicated, which is nonsensical.
    // Thus, we temporarily disable this->hasViews.
    from.settings_.hasViews = 0;
    initClone(from, GClone::ShareProps | GClone::Children);
    from.settings_.hasViews = 1;
    views->views_.push_back(this);
    settings_.hasViews = 1;
    return;
  }

  Gadget::initClone(from, c);

  if (c & GClone::Children) {
    if (auto* fromb = from.toBox()) {  // should always be verified
      for (auto& child : fromb->children_) {
        if (child) add(child->clone(c));
      }
      if (auto* blist = fromb->bchildren_) {
        for (auto& it : *blist)
          if (it.child_) borderAdd(it.child_->clone(c).toBox(), it.place_);
      }
    }
    // needed because bindings will be copied although p.get() != from.valueProp()
    if ((c & GClone::ShareProps) == 0) clearValues();
  }
}

void GBox::topInit() {
  if (omodes_.initialized) return;  // already done
  
  if (isTopLevel()) {
    init(nullptr);
  }
  else if (parent_ && parent_->omodes_.initialized) {
    Gadget *g{this}, *toplevel{};
    while (g) {
      if (g->isTopLevel()) {toplevel = g; break;}
      g = g->parent_;
    }
    if (toplevel) toplevel->init(nullptr);
    else error("topInit","internal error: toplevel gadget not found");
  }
}

void GBox::initChildren(GRender* parrender) {
  if (auto* blist = bchildren_) { // update border chidren
    for (auto& it : *blist) {
      it.child_->init(parrender);
    }
  }
  for (auto& it : children()) {    // update chidren
    it->init(parrender);
  }
}

void GBox::moveImpl(float x, float y, bool callcb) {
  if (frame_.x == x && frame_.y == y) return;
  auto dx = x - frame_.x;
  auto dy = y - frame_.y;
  frame_.x = x;
  frame_.y = y;

  for (auto& ch : children_) {
    ch->moveImpl(ch->frame_.x + dx, ch->frame_.y + dy, false);
  }

  if (auto* blist = bchildren_) {
    for (auto& b : *blist) {
      b.child_->moveImpl(b.child_->frame_.x + dx, b.child_->frame_.y + dy, false);
    }
  }

  if (callcb && (eventmask_ & GEventType::Move) != 0) {
    GEvent e(GTrigger::move, GCore::elapsedTime());
    doFire(e);
  }
  if (layout_.passivePos) {
    if (auto* pos = findProp<GPos>()) pos->set(frame_.x, frame_.y);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GBox::addProp(GProp& prop, bool add_to_list) {
  switch (prop.role()) {

    case GPropType::Role::Text:
      if (prop.isMutable()) *this << (TextField() << prop);
      else *this << (*new GLabel() << prop);
      return;
      
    case GPropType::Role::Expr:
      *this << (*new GLabel() << prop);
      return;
      
    case GPropType::Role::ExprWrapper:
      if (auto* e = prop.toExpr()) {
        *this << (*new GLabel() << *new GExprWrapper(*e));
      }
      return;
      
    case GPropType::Role::Number:
      if (prop.isMutable()) *this << (TextField() << prop);
      else *this << (*new GLabel() << prop);  // ou text pas editable ?
      return;

    case GPropType::Role::Bool:
      if (prop.isMutable()) *this << (CheckItem() << prop);
      else *this << (*new GLabel() << prop);   // ou check pas editable?
      return;

    case GPropType::Role::Choice:
      if (auto* c = prop.toChoice()) {
        setChoiceImpl(*c, add_to_list);
        c->fireDefaultCallbacks();
        updateLayout();
        /*
        modes_.hasChoice = 1;
        // should be addPropImpl(prop, add_to_list);
        props_.addReplace(prop, this);
        // updates box when changed
        c->startNotify(*this, [c](auto& g){c->radioCB(*g.toBox());});
        c->fireDefaultCallbacks();
        updateLayout();
         */
      }
      break;
      
    default:
      Gadget::addProp(prop, add_to_list);
      return;
  }
}

void GBox::setChoiceImpl(GChoice& c, bool add_to_list) {
  modes_.hasChoice = 1;
  //choice_ = c;
  addProp2(c, add_to_list);
  // updates box when changed
  c.startNotify(*this, [&c](auto& g){c.radioCB(*g.toBox());});
  for (auto& it : children_) {
    if (it->toItem()) it->setCheckable();
  }
}

void GBox::setValue(GString const&) {
  error("setValue","warning: This method does nothing (a box has no value");
}

void GBox::clearValues() {
  if (hasValue()) setValue("");
  for (auto& it : children()) it->clearValues();
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Gadget& GBox::createDefaultItem(GString const& arg) {
  return *new GLabel(arg);
}

void GBox::addItem(GItemSpec const& spec) {
  GString defclass;
  Gadget* item = addItemImpl(spec.spec_, -1, defclass);

  for (auto& it : spec.props_) {
    if (it->type() != GadgetHolder::Type()) {
      item->add(it);
    }
    else if (auto* holder = static_cast<GadgetHolder*>(it.get())) {
      if (auto* m = holder->gadget_->toMenu()) item->add(m);
      else if (auto* w = holder->gadget_->toWindow()) item->add(w);
    }
  }
}

void GBox::add(GString const& item) {
  GString defclass;
  addItemImpl(item, -1, defclass);
}

void GBox::add(GString const& item, int index) {
  GString defclass;
  addItemImpl(item, index, defclass);
}

void GBox::add(GStrings const& list) {
  GString defclass;
  for (auto& it : list) addItemImpl(it, -1, defclass);
}


Gadget* GBox::addItemImpl(GString const& item, int index, GString& defclass) {
  // search specific class for this item
  size_t pos{};
  GString cname = GItems::parseClass(item, pos);
  Gadget* g{};

  // specific class defined, will be used for next items if not redefined
  if (pos > 0) {
    if ((g = GadgetType::createInstance(cname, item.substr(pos)))) {
      defclass = cname;
      add(g, index);
      return g;
    }
    else error("addItem","gadget class not found: "+cname);
  }

  // default class is defined
  if (!defclass.empty()) {
    g = GadgetType::createInstance(defclass, item);
    add(g, index);
    return g;
  }

  // default class for items added directly to the gadget
  GItems* it{};
  if (settings_.hasItems && (it = findProp<GItems>()) && !it->itemclass_.empty()) {
    if ((g = GadgetType::createInstance(it->itemclass_, item))) {
      defclass = it->itemclass_;
      add(g, index);
      return g;
    }
    else error("addItem","gadget class not found: "+it->itemclass_);
  }

  g = &createDefaultItem(item);   // finally, default for this gagdget class
  add(g, index);
  return g;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int GBox::childCount() const {
  return (int)children().size();
}

Gadget* GBox::findChild(GadgetType& t, bool inborder) const {   // !!!!!!xxx
  if (inborder) {
    if (bchildren_) {
      for (auto& it : *bchildren_) {
        if (t == it.child_->type()) return it.child_;
      }
    }
  }
  else {
    for (auto& it : children_) {
      if (t == it->type()) return it;
    }
  }
  return nullptr;
}

Gadget* GBox::childAt(int pos) const {
  if (pos >= 0) {
    return (size_t(pos) >= children_.size()) ? nullptr : children_[pos];
  }
  else if (pos == -1) {
    return children_.empty() ? nullptr : children_.back();
  }
  else if (bchildren_) {
    for (auto& it : *bchildren_) {
      if (int(it.place_) == pos) return it.child_;
    }
  }
  return nullptr;
}

Gadget& GBox::operator[](int index) const {
  auto* g = childAt(index);
  if (g) return *g;
  else throw GFatalError("GBox::operator[]","invalid index: " + to_string(index));
}

int GBox::childPos(Gadget* child) const {
  if (!child) return -1;
  int pos = 0;
  for (auto& it : children_) {
    if (it.get() == child) return pos;
    ++pos;
  }
  return -1;

  //                                        !!!!!!! a_completer;
}

int GBox::borderPos(Gadget& child) const {
  if (!bchildren_) return -1;
  int ix = 0;
  for (auto& it : *bchildren_) {
    if (it.child_.get() == &child) return ix;
    ++ix;
  }
  return -1;
}

bool GBox::contain(Gadget* child, bool rec) const {
  if (!child || !child->parent_) return false;
  if (child->parent_ == this) return true;
  if (!rec) return false;

  Gadget* p = child->parent_->parent_;
  while (p) {
    if (p == this) return true;
    p = p->parent_;
  }
  return false;

  //                                         !!!!! a_completer;
}

Gadget* GBox::childAtPoint(GPoint const& pt, bool rec,
                           std::function<bool(Gadget&)> const& pred) {
  GPoint pos(pt);
  pos.x += this->frame_.x;
  pos.y += this->frame_.y;
  return childAtWPoint(pos, rec, pred);
}

// Note: si this est partiellement caché, ca peut renvoyer un child caché
Gadget* GBox::childAtWPoint(GPoint const& pos, bool rec,
                            std::function<bool(Gadget&)> const& pred) {
  Gadget* found{};
  GPickMode pick;
  pick.find_ = true;

  if (!pred) {
    if (!rec) {
      pick.pred_ = [this](Gadget& g){return &g == this || g.getParent() == this;};
    }
  }
  else {  // with pred
    if (rec) {
      pick.pred_ = [&](Gadget& g){return &g == this || (pred)(g);};
    }
    else {
      pick.pred_ = [&](Gadget& g){
        return &g == this || (g.getParent() == this && (pred)(g));
      };
    }
  }

  found = this->doPick(pos, pick);
  if (found == this) return nullptr;
  else return found;
}


int GBox::nearestChildIndex(GPoint const& pt, bool localpos) {
  if (children_.empty()) return 0;
  GPoint pos(pt);
  if (localpos) {
    pos.x += this->frame_.x;
    pos.y += this->frame_.y;
  }
  double dmin = std::numeric_limits<double>::max();
  int ix = 0, imin = 0;
  for (auto& it : children_) {
    double dx = pos.x - it->frame_.x;
    double dy = pos.y - it->frame_.y;
    double d2 = dx * dx + dy * dy;
    if (d2 < dmin) {dmin = d2; imin = ix;}
    ++ix;
  }
  auto* last = children_.back().get();
  double dx = pos.x - last->frame_.x;
  double dy = pos.y - last->frame_.y;
  if (isColumn()) dy -= last->frame_.h; else dx -= last->frame_.w;
  double d2 = dx * dx + dy * dy;
  if (d2 < dmin) {/*dmin = d2;*/ imin = ix;}
  return imin;
}

void GBox::perform(std::function< void(Gadget&)> fun) {
  for (auto& it : children()) (fun)(*it);
}

void GBox::perform(std::function< void(Gadget&, int pos)> fun) {
  int pos = 0;
  for (auto& it : children()) (fun)(*it, pos++);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GBox::borderAdd(Gadget* child, Place pos) {
  if (!child) {
    error("borderAdd","Can't add null child");
    return false;
  }
  if (!bchildren_) bchildren_ = new BorderList;

  auto it = bchildren_->begin();
  for (; it != bchildren_->end(); ++it) {
    if (it->place_ == pos) break;
  }

  if (it == bchildren_->end()) {
    bchildren_->push_back(BorderChild(child, pos));
    it = bchildren_->begin() + bchildren_->size() - 1;
  }
  else {
    if (it->child_) it->child_->parent_ = nullptr;
    it->child_ = child;
  }

  child->parent_ = this;
  child->onAdd(this);

  if ((eventmask_ & GEventType::AddRemove) != 0) {
    GEvent e(on::add, GCore::elapsedTime());
    e.aux_ = child;
    doFire(e);
  }

  return true;
}


void GBox::add(Gadget* child, int ix) {
  if (!child) {
    error("add","Can't add null child");
    return;
  }
  if (child == this) {
    error("add","Can't add gadget to itself");
    return;
  }

  if (auto* m = child->toMenu()) {
    Gadget::add(m);
  }
  else if (auto* w = child->toWindow()) {
    Gadget::add(w);
  }
  else if (child->place() != Place::Content) {
    auto place = child->place();
    if (place != Place::ScrollBar) borderAdd(child, place);
    else {
      if (child->isColumn()) borderAdd(child, Place::Right);
      else borderAdd(child, Place::Bottom);
    }
  }
  else {
    if (ix < 0 || size_t(ix) > children_.size()) ix = (int)children_.size();
    int oldpos = -1;
    
    // avoids deleting immediately the child if removed
    gptr<Gadget> sav;
    GBox* par = child->getParent();
    
    if (par == this) {    // moving child in same parent
      oldpos = childPos(child);
      if (oldpos == ix) return; // true;  // same pos, nothing to do
      else {
        sav = child;
        removeAt(oldpos);
        if (oldpos < ix) ix--;
      }
    }
    else if (par) { // other parent
      sav = child;
      par->remove(child);
    }
    
    child->parent_ = this;
    child->onAdd(this);
    
    if (size_t(ix) >= children_.size()) children_.emplace_back(child);
    else children_.insert(children_.begin()+ix, child);
    
    if (modes_.hasChoice) {
      if (auto* choice = choiceProp()) {
        choice->addBoxChild(*this, child, (ix >= 0 ? ix : childCount()-1));
      }
    }

    if ((eventmask_ & GEventType::AddRemove) != 0) {
      GEvent e(on::add, GCore::elapsedTime());
      e.aux_ = child;
      doFire(e);
    }

    // doit être apres insert
    updateLayout(Update::Css);
    
    // this objects has several views => clone gadgets that were added to it
    if (settings_.hasViews) {
      if (GBoxViews* views = findProp<GBoxViews>()) {
        for (auto* v : views->views_) {
          if (v != this) {
            v->settings_.hasViews = false;  // to avoid infinite loops
            *v << child->clone(GClone::ShareProps | GClone::Children);
            v->settings_.hasViews = true;
          }
        }
      }
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool GBox::remove(Gadget* child) {
  if (!child) {
    error("remove","Can't remove null child");
    return false;
  }
  
  /*
   if (child->toMenu()) {               TODO !!!!!!!!!
   return removeMenu(child->toMenu());
   }
   else if (child->toWindow()) {
   return removeWindow(child->toWindow());
   }
   */

  if (child->place() == Place::Content) {                       // xxxxxx !!!!
    auto it = std::find(children_.begin(), children_.end(), *child);
    if (it == children_.end()) return false;

    child->onRemove(this);

    if ((eventmask_ & GEventType::AddRemove) != 0) {
      GEvent e(GTrigger::remove, GCore::elapsedTime());
      e.aux_ = child;
      doFire(e);
    }
    child->parent_ = nullptr;
    children_.erase(it);
  }

  else if (!bchildren_) return false;

  else {
    int ix = borderPos(*child);
    if (ix < 0) return false;

    child->onRemove(this);

    if ((eventmask_ & GEventType::AddRemove) != 0) {
      GEvent e(GTrigger::remove, GCore::elapsedTime());
      e.aux_ = child;
      doFire(e);
    }
    child->parent_ = nullptr;
    bchildren_->erase(bchildren_->begin() + ix);
  }

  updateLayout();
  return true;
}

/*
int GBox::borderPos(int place) const {
  if (!bchildren_) return -1;
  int ix = 0;
  for (auto& it : *bchildren_) {
    if (int(it.place_) == place) return ix;
    ++ix;
  }
  return -1;
}*/

bool GBox::removeAt(int pos) {
  if (pos >= -1) {  // content area

    int len = int(children_.size());
    if (pos == -1) pos = len - 1;
    if (pos < 0 || pos >= len) {
      //throw GFatalError("GBox::removeChild","invalid position: " + to_string(index));
      error("removeAt","invalid position: " + to_string(pos));
      return false;
    }
    auto it = children_.begin() + pos;
    (*it)->onRemove(this);

    if ((eventmask_ & GEventType::AddRemove) != 0) {
      GEvent e(GTrigger::remove, GCore::elapsedTime());
      e.aux_ = *it;
      doFire(e);
    }
    (*it)->parent_ = nullptr;
    children_.erase(it);
  }

  else if (!bchildren_) return false;

  else {   // border area
    auto it = bchildren_->begin();
    for (; it != bchildren_->end(); ++it) {
      if (int(it->place_) == pos) break;
    }
    if (it == bchildren_->end()) return false;

    it->child_->onRemove(this);

    if ((eventmask_ & GEventType::AddRemove) != 0) {
      GEvent e(GTrigger::remove, GCore::elapsedTime());
      e.aux_ = it->child_;
      doFire(e);
    }

    it->child_->parent_ = nullptr;
    bchildren_->erase(it);

  }
  updateLayout();
  return true;
}


void GBox::cleanup() {
  removeChildren(Remove(Remove::InContent | Remove::InBorder | Remove::Internals));
  Gadget::cleanup();
}

void GBox::removeChildren(Remove mode) {

  if ((mode & Remove::InContent) != 0) {
    GadgetList internals;

    for (auto& c : children_) {
      if ((mode & Remove::Internals) == 0 && c->isInternal()) {
        internals.push_back(c);
      }
      else {
        c->onRemove(this);

        if ((eventmask_ & GEventType::AddRemove) != 0) {
          GEvent e(GTrigger::remove, GCore::elapsedTime());
          e.aux_ = c;
          doFire(e);
        }
        c->parent_ = nullptr;
      }
    }
    children_.clear();
    if (!internals.empty()) children_ = internals;
  }

  if ((mode & Remove::InBorder) != 0 && bchildren_) {
    for (auto& it : *bchildren_) {
      auto* c = it.child_.get();
      c->onRemove(this);

      if ((eventmask_ & GEventType::AddRemove) != 0) {
        GEvent e(GTrigger::remove, GCore::elapsedTime());
        e.aux_ = c;
        doFire(e);
      }
      c->parent_ = nullptr;
    }
    delete bchildren_;
    bchildren_ = nullptr;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


void GBox::doPaint(GRenderGraphics& g, GRender* parr) {
  if (modes_.checked && layout_.reshapeFrame) {
    GRender r(*this, parr);
    r.paintBox(g, *this);
    paintHandleFrame(g);
  }
  else {
    GRender(*this, parr).paintBox(g, *this);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GVBox& VBox(GString const& s) {return *new GVBox(s);}


GButtonBox::GButtonBox(GadgetType& t, GString const& spec)
: GBox(t, spec) {}

GButtonBox::GButtonBox(GString const& spec)
: GBox(Type(), spec) {}

Gadget& GButtonBox::createDefaultItem(GString const& arg) {
  return *new GItem(arg);
}


}
