//
// Charts
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Chart_hpp
#define Guit_Chart_hpp
#include <gchoicebox.hpp>
#include <gvalueitem.hpp>
namespace guit {

/// Bar or Pie Chart.
/// A ChartBox should normaly contain GChartItem items. This is the type that is used by default
/// when a string is added to a ChartBox (the string is implicitely converted into a GChartItem)
///
/// Important: a ChartBox must have a size (see GSize) and/or an appropriate flex value (see GFlex),
/// because it can't deduce its size form the items it contains (as they adapt to the ChartBox)
/// For the same reason, setRange() should be called to specify the min and max values of the items
/// (the default min and max values are 0 and 100)
///
/// The defaut orientation is 'row'. This can be changed through the gadget's flow prop (see GFlow):
/// - row : items are displayed in a row (items are thus vertical)
/// - colum : items are displayed in a column (items are horizontal)
/// - pie : items are displayed as in a pie chart
///
class GChartBox : public GChoiceBox {
protected:
  GChartBox(GadgetType&, GString const& css_and_title);

public:
  GChartBox(GString const& css_and_title = "");
  GChartBox& clone(GClone const&) override;
  GChartBox* toChartBox() override {return this;}

  virtual GChartBox& setRange(float min, float max);
  float min() const {return min_;}
  float max() const {return max_;}

  using GBox::add;
  void add(Gadget* child, int pos = -1) override;
  void add(Gadget& child, int index = -1) override {GChartBox::add(&child, index);}

  using GType = GadgetType_<GChartBox,GChoiceBox>;
  static GType& Type();
  GType& type() const override {return Type();}

  Gadget& createDefaultItem(GString const& arg) override;
  void initClone(Gadget& from, GClone const&) override;

  /// @internal
  float min_{0.f}, max_{100.f};
};


/// Creates a GChart (Bar or Pie Chart).
GChartBox& ChartBox(GString const& css_and_title = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


/// Items for charts.
/// A ChartItem is typically added to a GChart.
/// - A GNumber prop (e.g. GFloat or GInt) can be added to this gadget to control its value.
/// - A GText prop can be added to this gadget to control the text it displays. The color of the
///  changes automatically depending on the GBackground color of the gadget (white if the
///  background color is dark, black otherwise)
/// - By default, chart items are checkable (see Gadget::setCheckable()) and their value can
///  be changed interactively by dragging the mouse.
class GChartItem : public GValueItem {
  GChartItem(GadgetType&, GString const& css_value_and_title);

public:
  GChartItem(GString const& css_value_and_title = "");
  GChartItem(float value);
  
  GChartItem& clone(GClone const&) override;

  using GType = GadgetType_<GChartItem,GValueItem>;
  static GType& Type();
  GType& type() const override {return Type();}
};


/// Creates a GChartItem (Items for charts).
/// @{
GChartItem& ChartItem(GString const& css_value_title = "");
GChartItem& ChartItem(float value);
/// @}

}
#endif
