//
//  Dialog boxes.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Dialog_hpp
#define Guit_Dialog_hpp
#include <gwindow.hpp>
#include <gbool.hpp>
namespace guit {
  
/// Dialog box (specialization of GWindow).
/// Optionally, a Dialog can contain OK, Close and other buttons:
/// see okButton(), cancelButton(), setButtons().
///
/// This class also provides static methods for opening predefined dialogs: see alert() and confirm().
///
/// ### Key methods:
/// - setModal(state) : makes the window modal or not (not modal by default)
/// - open() makes the dialog modal, then shows the dialog and blocks until the OK, Cancel
///  button of the dialog or the close button of the dialog bar is clicked (calls close())
/// - show(state) : shows/hides the window. show(false) does not call close()
/// - close() : closes the window. Called when the OK, Cancel button of the dialog
///  or the close button of the dialog bar is clicked
/// - dieOnClose(state) : behavior when close() is called (see GWindow).
///
/// ### Triggers and callbacks:
/// - See GWindow
///
/// ### Closing the dialog:
/// - See GWindow
///
class GDialog : public GWindow {
protected:
  GDialog(GadgetType&, GString const& css_and_title);

public:
  GDialog(GString const& css_and_title = "");
  GDialog& clone(GClone const&) override;
  
  /// Open an Alert dialog with one button.
  /// Opens a predefined dialog with a OK button that blocks until the user clicks the button.
  /// @return Always returns 0.
  static int alert(GString const& message, GString const& info = "");
  
  /// Open an Alert dialog with two buttons.
  /// Opens a predefined dialog with a Cancel and a OK button that  blocks until the user clicks one of the buttons.
  /// @return the number of the clicked button, ie 0 for Cancel and 1 for OK. Also returns 0 if the dialog
  /// is closed by clinking the close button of the window title bar.
  static int confirm(GString const& message, GString const& info = "",
                     GString const& buttons = "");

  /// Shows/hides the dialog.
  /// The dialog is modal if setModal(true) was called.
  /// @note contrary to open(), this function is non-blocking.
  GDialog& show(bool state = true) override;
  
  /// Shows the dialog and blocks.
  /// Opens the dialog and blocks until the user clicks a button.
  /// This function makes the dialog modal: contrary to show() it blocks the interaction flow.
  /// @return 0 if Cancel is clicked, 1 if OK is clicked, and so on (see result()).
  virtual int open();
  
  /// Returns the number of the button that was clicked.
  /// Returns the number of the button in the buttonBox(). By convention, 0 is the Cancel button
  /// and 1 the OK button.
  int result() const {return result_;}
  
  /// Returns/creates the Cancel button.
  /// Creates this button if needed. The Cancel button is the first button in the buttonBox().
  /// result() will return 0 when this button is clicked.
  virtual Gadget& cancelButton();
  
  /// Returns/creates the OK button (and creates a Cancel button if neeeded).
  /// Creates the Cancel and OK buttons if needed, returns the OK button.
  /// The OK button is the second button in the buttonBox().
  /// result() will return 1 when this button is clicked.
  virtual Gadget& okButton();
  
  /// Returns/creates the box that contains the buttons.
  /// Creates the box if needed. Additional buttons can be added to this box. By convention,
  /// the first button is the Cancel button (number 0), the second button is the OK button, etc.
  virtual GBox& buttonBox();
  
  /// Add/Changes buttons.
  /// Remove previous buttons and creates as many buttons as there are tokens
  /// separated by commas in _buttons_.  The first token must be the name of the Cancel button,
  /// the second token (if any) the name of the OK button, etc.
  /// result() will return the number of the corresponding button (starting at 0 for Cancel).
  virtual void setButtons(GString const& buttons);
  
  /// Returns the button at this position.
  /// Returns null if there is no gadget at this position or if it is not a GItem.
  virtual GItem* findButton(int pos);

  using GType = GadgetType_<GDialog,GWindow>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  gptr<GBox> bbox_;
  int result_{};
  bool quitloop_{};
  virtual void setResult(int result, bool hide);
};

/// Creates a Dialog
/// The argument is the title of the dialog.
/// @see GDialog.
GDialog& Dialog(GString const& css_and_title = "");


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Dialog for showing a message and/or entering text.
/// By default, an InfoDialog is modal (see setModal()) and has a Cancel and a OK button.
///
/// The dontOpenIf() method allows performing the OK action without opening the dialog.
/// This feature is particularily useful with a QuitDialog: it makes it possible to quit the application
/// whithout opening the QuitDialog when there is nothing to save.
///
/// @see GDialog.
class GInfoDialog : public GDialog {
protected:
  GInfoDialog(GadgetType&, GString const& css_and_title);

public:
  GInfoDialog(GString const& css_and_title = "");
  GInfoDialog& clone(GClone const&) override;

  GInfoDialog& show(bool state = true) override;

  /// Returns a label showing a message.
  /// Its properties can be set as follows:
  /// @code
  ///   mydialog.message() << "<color:red><font:italic>Do you want to quit?"
  /// @endcode
  GItem& message();
  
  /// Returns/creates a label for displaying complementary information.
  /// see message() for details.
  GItem& info();
  
  /// Returns/creates an input textfield.
  GTextItem& input();
  
  /// Open the dialog only if a condition is true (useful for QuitDialog).
  ///@{
  GInfoDialog& openOnlyIf(GBool& test);
  GInfoDialog& openOnlyIf(GExprBool const& test);
  GInfoDialog& openOnlyIf(std::function<bool()> test);
  ///@}

  using GType = GadgetType_<GInfoDialog,GDialog>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  gptr<GItem> message_, info_;
  gptr<GTextItem> input_;
  gprotect<GProp> openonly_;
};

/// Creates an InfoDialog.
GInfoDialog& InfoDialog(GString const& title="", GString const& message = "");

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


/// Quit Dialog.
/// When a QuitDialog  is added to a Window (or another Dialog), it is automatically opened
/// when the Window/Dialog is closed (see GWindow::close()).
/// Moreover, its Quit button quits the application.
/// @code
///     // quits without showing the dialog if changed is false
///     auto& quitdial = QuitDialog().openOnlyIf(changed);
/// @endcode
/// @see GInfoDialog.
class GQuitDialog : public GInfoDialog {
protected:
  GQuitDialog(GadgetType&, GString const& css_and_title);

public:
  GQuitDialog(GString const& css_and_title = "");

  GQuitDialog& clone(GClone const&) override;

  using GType = GadgetType_<GQuitDialog,GInfoDialog>;
  static GType& Type();
  GType& type() const override {return Type();}
};

/// Creates a QuitDialog.
GQuitDialog& QuitDialog(GString const& title="", GString const& message="");

}
#endif

