//
// Shape Gadgets (rectangles, lines, ovals, etc.)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Shape_hpp
#define Guit_Shape_hpp
#include <vector>
#include <gadget.hpp>
#include <ggraphics.hpp>
namespace guit {

class GShape : public Gadget {
protected:
  friend GadgetType_<GShape,Gadget>;
  GShape(GadgetType&, GString const& css);
  GShape(GString const& css);

public:
  ~GShape();
  Gadget& clone(GClone const&) override;

  GShape* toShape() override {return this;}

  using Gadget::findProp;
  GProp* findProp(GPropType&) const override;

  using Gadget::prop;
  GProp& prop(GPropType&, bool compute_style = false) override;

  /// Returns the prop controlling the position of the GShape.
  /// Returns the same prop as prop<GPos>().
  /// This prop can be changed just by adding a new GPos to the GShape, ie:  shape << new GPos(x,y);
  GPos& posProp();

  /// Returns the prop controlling the size of the GShape.
  /// Returns the same prop as prop<GSize>().
  /// This prop can be changed just by adding a new GSize to the GShape, ie:  shape << new GSize(w,h);
  GSize& sizeProp();

  void move(float x, float y) override;
  void move(GPoint const& p) override {move(p.x, p.y);}

  void resize(float w, float h) override;
  void resize(GDim const&) override;

  using GType = GadgetType_<GShape,Gadget>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @ internal
  /// @{
  gptr<GPos> pos_;
  gptr<GSize> size_;
  void onAdd(GBox* parent) override;
  void addProp(GProp& prop, bool add_to_list) override;
  /// @}
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Rectangle.
class GRectShape : public GShape {
protected:
  GRectShape(GadgetType&, GString const& css);

public:
  GRectShape(GString const& css = "");
  GRectShape& clone(GClone const&) override;

  using GType = GadgetType_<GRectShape,GShape>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @ internal
  /// @{
  void doPaint(GRenderGraphics&, GRender* parent_render) override;
  /// @}
};

/// Creates a Rectangle.
GRectShape& RectShape(GString const& title = "");

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Circle or Ellipse.
class GOvalShape : public GRectShape {
protected:
  GOvalShape(GadgetType&, GString const& css);

public:
  GOvalShape(GString const& css = "");
  GOvalShape& clone(GClone const&) override;

  using GType = GadgetType_<GOvalShape,GRectShape>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @ internal
  /// @{
  void doPaint(GRenderGraphics&, GRender* parent_render) override;
  Gadget* doPick(GPoint const&, GPickMode&) override;
  /// @}

};

/// Creates a Circle or an Ellipse.
GOvalShape& OvalShape(GString const& title = "");

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Arc or Pie.
class GArcShape : public GOvalShape {
protected:
  GArcShape(GadgetType&, GString const& css);

public:
  GArcShape(GString const& css = "");
  GArcShape& clone(GClone const&) override;

  GArcShape& setStart(float start);
  GArcShape& setExtent(float extent);
  float start()  const;
  float extent() const;
  
  GFloat& startProp()  {return *start_;}
  GFloat& extentProp() {return *extent_;}

  using GType = GadgetType_<GArcShape,GOvalShape>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @ internal
  /// @{
  gptr<GFloat> start_, extent_;
  void doPaint(GRenderGraphics&, GRender* parent_render) override;
  Gadget* doPick(GPoint const&, GPickMode&) override;
  /// @}
};

/// Creates an Arc or a Pie.
GArcShape& ArcShape(GString const& title = "");

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Line Shape.
/// fromPos() and toPos() are the points specifying the beginning and the end of the line.
/// Note that findProp<GPos>() returns fromPos() .
class GLineShape : public GShape {
protected:
  GLineShape(GadgetType&, GString const& css);

public:
  GLineShape(GString const& css = "");
  GLineShape& clone(GClone const&) override;

  /// Returns the prop controlling the ending position of the line.
  /// inherited method posProp() returns the starting position, which is also the prop returned
  /// by prop<GPos>().
  /// @{
  GPos& pos2Prop();
  /// @}

  /// Changes the props controlling the starting and ending position of the line.
  /// changePosProp() is equivalent to adding a GPos to the line ie: line << new GPos(x,y);
  /// @{
  GLineShape& changePosProp(GPos& pos);
  GLineShape& changePos2Prop(GPos& pos);
  /// @}

  void move(GPoint const& p) override {move(p.x, p.y);}
  void move(float x, float y) override;

  void reshapeInteractively() override;
  void reshapeInteractively(GMouseEvent& e) override;

  using GType = GadgetType_<GLineShape,GShape>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @ internal
  /// @{
  gptr<GPos> pos2_;
  void onAdd(GBox* parent) override;
  void addProp(GProp& prop, bool add_to_list) override;
  void doPaint(GRenderGraphics&, GRender* parent_render) override;
  Gadget* doPick(GPoint const&, GPickMode&) override;
  void packPos(float parentx, float parenty) override;
  void moveImpl(float x, float y, bool callcb) override;
  virtual void updateFrame();
  GHandle handleDown(GPoint const& wpos, bool inter) override;
  void handleDrag(float dx, float dy, GHandle const&) override;
  void handleUp(GPoint const& wpos, GHandle const&) override;
  GHandle findHandle(GPoint const& wpos) override;
  void readAux(std::istream&, class GIOBuffer&) override;
  void writeAux(std::ostream&, class GIOBuffer&) const override;
  /// @}
};

/// Creates a Line.
GLineShape& LineShape(GString const& title = "");

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Free hand drawing.
class GFreehandShape : public GShape {
protected:
  GFreehandShape(GadgetType&, GString const& css);

public:
  GFreehandShape(GString const& css ="");
  GFreehandShape& clone(GClone const&) override;

  void addPoint(float x, float y) {addPoint(GPoint(x, y));}
  void addPoint(GPoint const&);
  GFreehandShape& operator<<(GPoint const& p) {addPoint(p); return *this;}
  void getPoints(GPoints&) const;

  using GType = GadgetType_<GFreehandShape,GShape>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @ internal
  /// @{
  GPoint pmin_, pmax_, last_;
  GPoints points_;
  void packPos(float parentx, float parenty) override;
  void moveImpl(float x, float y, bool callcb) override;
  GHandle handleDown(GPoint const& wpos, bool inter) override;
  void handleDrag(float dx, float dy, GHandle const&) override;
  void handleUp(GPoint const& wpos, GHandle const&) override;
  void doPaint(GRenderGraphics&, GRender* parent_render) override;
  void readAux(std::istream&, class GIOBuffer&) override;
  void writeAux(std::ostream&, class GIOBuffer&) const override;
  /// @}
};

/// Creates a PolyLine or a Polygon.
GFreehandShape& FreehandShape(GString const& title = "");

}
#endif
