//
//  Base class for Windows and Dialogs.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gapp.hpp>
#include <gtrigger.hpp>
#include <gtext.hpp>
#include <gicon.hpp>
#include <ghotkey.hpp>
#include <gwindow.hpp>
#include <gdialog.hpp>
#include <gitem.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(Window)

GWindow::GWindow(GadgetType& t, GString const& spec) : GBox(t) {
  wmodes_.mustupdate = true;
  wmodes_.dieonclose = false;  // !!!!

  if (!wmodes_.softwin) {
    natwin_ = new GNatWindow(this, true, true);
    GCore::addWindow(*this);
  }
  *this << new GPos();  // nb: modif api pour move !

  // dont call GBox(t,spec) otherwise title displayed inside window
  if (!spec.empty()) {
    GString title;
    GadgetStyle::initStyle(*this, spec, title);
    if (!title.empty()) *this << new GTitle(title);
    //if (!title.empty()) setTitle(title);
  }
}

GWindow::GWindow(GString const& spec)
: GWindow(Type(), spec) {}
 
GWindow::~GWindow() {
  //cerr << "--- ~GWindow() " << this << endl;
  if (!wmodes_.softwin) {
    GCore::removeWindow(*this);
    delete natwin_;
  }
  delete hotkeys_;
}

GWindow& GWindow::clone(GClone const& cm) {
  auto* g = new GWindow;
  g->initClone(*this, cm);
  return *g;
}

int GWindow::startApp() {
  return GCore::start(this);
}

void GWindow::topInit() {
  if (omodes_.initialized) return;  // already done
  GBox::topInit();
  GIcon* icon{};
  if (style_ && (icon = style_->findProp<GIcon>(*this))) {
    setIconImpl(icon->value().filename);
  }
}

void GWindow::setModal(bool modal) {
  wmodes_.modal = modal;
}

void GWindow::add(GWindow* opened) {
  if (!opened) error("add","Can't add null child");
  else {
    if (opened->to<GQuitDialog>()) addOpener(*opened, GTrigger::done, -1, true);
    else Gadget::add(opened);
  }
}


void GWindow::add(Gadget* child, int ix) {
  if (!child) {
    error("add","Can't add null child");
    return;
  }
  if (child->type() == GTitle::Type()) {
    //if (wmodes_.softwin) GBox::setTitle(title); else natwin_->setTitle(title);
    natwin_->setTitle(child->text());
  }
  else GBox::add(child, ix);
}

/*
 void GWindow::setTitle(GString const& title) {
 if (wmodes_.softwin) GBox::setTitle(title);
 else natwin_->setTitle(title);
 }
*/

void GWindow::setIconImpl(GString const& filename) {
  if (wmodes_.softwin) GBox::add(new GIcon(filename));
  else natwin_->setIcon(filename);
}

void GWindow::addProp(GProp& prop, bool add_to_list) {
  if (prop.role() == GPropType::Icon) {
    if (auto* i = dynamic_cast<GIcon*>(&prop)) {
      setIconImpl(i->value().filename);
    }
  }
  else GBox::addProp(prop, add_to_list);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GWindow::updateLayout(Update mode) {
  if (wmodes_.softwin) {
    GBox::updateLayout(mode);
    return;
  }
  if (!isShown() && mode == Update::IfNeeded) return;
  
  if (mode != Update::Now) {
    GCore::updateLayout(this, nullptr, mode == Update::Css);
  }
  else {
    float w = frame_.w, h = frame_.h;
    computeSize(nullptr, true);  // parent_render is null for Window
    pack(nullptr);

    if (frame_.w != w || frame_.h != h) {
      // NOTE: calls Gadget::resize(), which modifies GSize
      natwin_->resize(frame_.w, frame_.h);
    }
    wmodes_.mustlayout = false;
    update();
  }
}

GWindow& GWindow::show(bool state) {
  if (wmodes_.softwin) {
    GBox::show(state);
    return *this;
  }
  if (state == modes_.shown) {
    if (state) {restore(); raise();}   // always raise the window if state is true
    return *this;
  }
  modes_.shown = state;
  
  if (state) {
    if (!omodes_.initialized) topInit();
    updateLayout(Update::Now);
  }
  
  if (wmodes_.modal) GCore::setModalWindow(state ? this : nullptr);
  natwin_->show(state);
  if (state) {restore(); raise();}
  
  if ((eventmask_ & GEventType::Show) != 0) {
    GEvent e(state ? on::show : on::hide, GCore::elapsedTime());
    e.window_ = this;
    doFire(e);
  }
  return *this;
}

void GWindow::show(GEvent& e, bool autoplace) {
  parent_ = &e.from();
  //place_ = BelowParent;
  show(true);
}

void GWindow::close() {
  //cerr << "close " << int(wmodes_.quitcalled ) << endl;

  // when closing the window, calls on::done callbacks if any,
  // otherwise quits the app or destroys the window
  if (!wmodes_.quitcalled && (eventmask_ & GEventType::Done) != 0) {
    GEvent e(on::done, GCore::elapsedTime());
    e.window_ = this;
    doFire(e);
    if (!wmodes_.quitcalled && (!e.hasGuard() || e.guardDone())) return;
  }
  wmodes_.quitcalled = 0;
  
  if (this == GCore::mainWindow()) {
    //if (!GCore::modalWindow()) {
      //cerr << "exit " << int(wmodes_.quitcalled ) << endl;
      GApp::exit(0);
    //}
  }
  else {
    show(false);
    // deletes the window if not referenced elsewhere
    if (wmodes_.dieonclose && !omodes_.plain) {gptr<GWindow> p = this;}
  }
}

void GWindow::dieOnClose(bool dies) {
  wmodes_.dieonclose = dies;
}

void GWindow::moveImpl(float x, float y, bool callcb) {
  if (wmodes_.softwin) {
    GBox::moveImpl(x, y, callcb);
  }
  // probleme avec ce test : la premiere fois !
  else {   //if (screenpos_.x != x || screenpos_.y != y) { n
    screenpos_.x = x;
    screenpos_.y = y;
    natwin_->move(x, y);
    if ((eventmask_ & GEventType::Move) != 0) {
      GEvent e(GTrigger::move, GCore::elapsedTime());
      doFire(e);
    }
  }
}

void GWindow::setScale(float scale) {
  scale_ = scale;
  update();
}

void GWindow::raise() {
  if (!wmodes_.softwin) natwin_->raise();
}

void GWindow::minimize() {
  if (!wmodes_.softwin) natwin_->minimize();
}

void GWindow::maximize() {
  if (!wmodes_.softwin) natwin_->maximize();
}

void GWindow::restore() {
  if (!wmodes_.softwin) natwin_->restore();
}

GWindow& GWindow::setBordered(bool bordered) {
  if (!wmodes_.softwin) natwin_->setBordered(bordered);
  return *this;
}

GWindow& GWindow::setResizable(bool r)  {
  if (!wmodes_.softwin) natwin_->setResizable(r);
  return *this;
}

GWindow& GWindow::setMinSize(float w, float h) {
  if (!wmodes_.softwin) natwin_->setMinSize(w, h);
  return *this;
}

GWindow& GWindow::setMaxSize(float w, float h) {
  if (!wmodes_.softwin) natwin_->setMaxSize(w, h);
  return *this;
}

GWindow& GWindow::setOpacity(float opacity) {
  if (!wmodes_.softwin) natwin_->setOpacity(opacity);
  return *this;
}

#if defined(__APPLE__)

GWindow& GWindow::stayOnTop() {
  natwin_->stayOnTop();
  return *this;
}

GWindow& GWindow::stayOnTopOf(GWindow& w) {
  natwin_->stayOnTopOf(*w.natwin_);
  return *this;
}

GWindow& GWindow::joinAllSpaces() {
  natwin_->joinAllSpaces();
  return *this;
}
#endif

uint32_t GWindow::windowID() const {
  return wmodes_.softwin ? 0 : natwin_->winID();
}

// incomplet il faudrait detecter qunad une fenetre passe au dessus !!!
GWindow* GWindow::findWindow(GPoint const& screenpos) {
  GWindow* found{};
  for (auto* w : GCore::windowList()) {
    /*
     if (w)
     cerr << w << " " << int(w->isShown())
     << " " << w->screenpos_.x
     << " " << w->screenpos_.y
     << " " << w->frame_.w
     << " " << w->frame_.h
     <<endl;*/
    if (w && w->isShown() && w->natwin_
        && screenpos.x >= w->screenpos_.x
        && screenpos.y >= w->screenpos_.y
        && screenpos.x < w->screenpos_.x + w->frame_.w
        && screenpos.y < w->screenpos_.y + w->frame_.h
        ) {
      found = w;  // on ne teste pas qui est dessus !!
      break;
    }
  }
  return found;
}

// note that update() does not produce an event (at least with SDL), it will
// just make SDL call paintCB() on this window after processing an event
void GWindow::update() {
  /*
  if (wmodes_.softwin) {
    GBox::update();
    return;
  }*/
  if (!isShown()) return; 
  wmodes_.mustupdate = true;
  //cerr << "\nGWindow::update "<< this <<endl;
  GCore::requestUpdate();
}

void GWindow::addHotkey(Gadget& g, GHotkey& h) {
  if (h.value_.key == 0) {
    error("addHotkey","invalid hotkey: " + h.text_);
    return;
  }
  //cerr << "addHotkey " << h.value_.state << " "<< h.value_.key<<endl;
  if (!hotkeys_) hotkeys_ = new GHotkeys;

  auto it = hotkeys_->find(h.mapcode_);
  if (it != hotkeys_->end()) {
    error("addHotkey","kotkey "+ h.text_+" already exists (ignored)");
  }
  else {
    // null means window; this is to avoid circular refs
    (*hotkeys_)[h.mapcode_] = (&g == this) ? nullptr : &g;
  }
}

bool GWindow::triggerHotkey(GEventState keystate, GKeyCode key) {
  if (!hotkeys_) return false;
  GHotkey k;
  k.value_.modifiers = keystate;
  k.value_.key = tolower(key);
  //cerr << " Hotkey " << k.value_.state << " "<<  k.value_.key<<endl;

  auto it = hotkeys_->find(k.mapcode_);
  if (it == hotkeys_->end()) return false;

  //cerr << "Found Hotkey " << it->first <<endl;
  // null means window; this is to avoid circular refs
  GItem* item{};
  if (it->second && (item = it->second->toItem())) {

    // check the item is still a child of this window
    if (item->getWindow() == this) item->doAction();
    else {
      // the item has been removed => remove the hotkey from the map
      item->settings_.hotkeySet = 0;
      hotkeys_->erase(it);
    }
  }
  return true;
}

}
