//
// Application context (must be created first)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#ifndef GuitApp_hpp
#define GuitApp_hpp
#include <gtypes.hpp>
namespace guit {

/// Application Context (must be created first).
/// The app context must be created before creating any Guit object, as in the example below:
/// @code
///  #include <guit.hpp>
///  using namespace guit;
///
///  int main(int argc, const char **argv) {
///     GApp app(argc, argv);
///
///     /// allows inspecting gadgets when cmd-pressing the mouse on them
///     /// (or by calling showInspector())
///     app.inspector();
///
///     // loads CSS rules
///     app
///     << "#myapp {margin: 15 15}"
///     << ".mypanel {border: rounded; font: bold 16 arial,sans-serif}";
///
///     // loads CSS files
///     app << "myapp.css" << "whatever.css";
///
///     // starts the event loop, the window given as an argument (here an
///     // instance of a custom MyWindow class) become the 'main window'
///     return start(new MyWindow);
///  }
/// @endcode
/// CSSs should be loaded before creating the gadgets.
///
/// Files (CSSs, icons, etc.) that do not start with /  \  . and magic characters ~ : ! are searched in
/// the application resource directory (MacOSX only), then in guit-dir/exec-name where guit-dir is
/// the Guit directory of the user, and exec-name the name of the executable. guit-dir is located here:
/// - MacOS: ~/Library/Guit
/// - Linux: ~/.guit
/// - Windows: ~/Guit
///
/// Magic characters must be the first character of a file name:
/// - ~  is the home directory of the user
/// - !  is the Guit directory of the user (ie guit-dir)
/// - :  is the first resource directory (ie guit-dir/exec-name, except for MacOS applications where
/// it's the application resource directory, ie application/Content/Resources)
///
/// start() starts the event loop and blocks until the application is terminated. start()  returns a status
/// which is usually 0 (see quit() and exit()). Its argument must derive from GWindow and  should
/// be created in the heap (i.e. using new). This window will be the *main window* of the application.
/// By default, the application is terminated when this window is closed. This can be changed by
/// adding a GQuitDialog to it or on::done callbacks (see GWindow).
///
/// Example:
/// @code
///  class MyWindow : public GWindow {
///  public:
///    MyWindow() : GWindow("#myapp My Nice Window")
///    {
///       auto& panel = VBox(".mypanel")
///       << TextField("initial text")
///       << Button("Click Me")
///       << ScrollBar("50");
///
///       // add gadgets to the window
///       *this
///       << QuitDialog()
///       << panel
///       << ...;
///    }
///  };
/// @endcode
///
/// A CSS class  (.mypanel) and/or ID (#mywin) can be given as an argument to the gadget
/// constructors. They can be followed by text that will be the title or value of the gadget
/// (eg the title of a GWindow/GBox, the text of a GButton/GItem/GTextField, the value of
/// a GSlider/GChartItem)
///
/// VBox(), TextField(), Button(), etc. are shortcut functions that instantiate the corresponding class.
/// For instance, Button("Click me") is equivalent to:  * new GButton("Click Me").
/// Note that classes start with a 'G', contrary to shortcut functions.
///
/// Guit relies on TrueType fonts. A font list cache is created in the users' Guit directory the first type
/// a Guit application is launched. It should be removed when new fonts are added to the system,
/// so that it can be built again.
class GApp {
public:
  /// Inits the application context; must be called before creating any other object.
  GApp(int argc, const char** argv);

  /// Starts the event loop
  /// This function is blocking and must be called last.
  /// @{
  int start(GWindow& main_window);
  int start(GWindow* main_window);
  /// @}

  /// Closes or quits the application.
  /// - close() opens a GQuitDialog if one was added to the main window, otherwise it calls quit()
  /// - quit() and exit() quit the application unconditionally
  /// - start() or status() will return the value given as an argument to exit() (0 if quit() is called)
  /// @see GWIndow.
  /// @{
  static void close();
  static void quit() {exit(0);}
  static void exit(int status = 0);
  /// @}

  /// Returns the value given as an argument to exit().
  static int status();

  /// Adds CSS.
  /// @{
  static bool addCssFile(GString const& css_file);
  static bool addCssRules(GString const& css_rules);
  GApp& operator<<(GString const& css);
  ///@}

  /// Directories where files (CSS, icons, etc.) are searched by default.

  /// @{
  static GStrings const& resourceDirs();

  /// Returns the user's Guit directory.
  /// Returns:
  /// - ~/Library/Guit on MacOS
  /// - ~/Guit on Windows
  /// - ~/.guit on Linux (note the dot)
  static GString const& guitDir();

  /// Returns guitDir()/execName()
  /// The ! magic character refers to this directory.
  static GString const& appDir();

  /// Returns the name and the path of the executable.
  /// execPath() returns argv[0] and execName() its basename().
  /// @{
  static GString execName();
  static GString execPath();
  ///@}

  /// Returns the arguments of the command line.
  /// - argc() returns the number of arguments
  /// - argv(n) returns the n-th argument (empty string if n is out of range).
  /// @{
  static int argc();
  static GString argv(int n);
  ///@}

  /// Searches if one of the arguments of the command line is _name_.
  /// Note that arvg(0) is ignored.
  static bool hasArg(GString const& name);

  /// Searches if one of the arguments of the command line is _name_ or name=value.
  /// - returns the associated _value_ (empty string if no value)
  /// - _assign_ specifies the assignement sign(s), eg: if _assign_ is ":"  then : will be searched
  /// instead of = moreover both : and = will be searched is _assign_ is  "=:"
  /// - note that arvg(0) is ignored.
  static bool hasArg(GString const& name, GString& value,
                     GString const& assign = "=");

  /// Returns the size of the main screen.
  static GDim screenSize();

  /// Returns the Main Window.
  /// The main window is the argument that was provided to start().
  static GWindow* mainWindow();

  /// Returns the gadget having the focus (null otherwise).
  static Gadget* currentFocus();

  /// Requests all gadgets to be repainted.
  /// @See repaintAll().
  static void updateAll();

  /// Immediately repaints all gadgets.
  /// @See updateAll() .
  static void paintAll();

  /// Executes a function in the main thread.
  /// calls gexec().
  /// @warning Functions that modify/update the GUI must always be called in the main thread.
  static void exec(GFunction const& fun);

  /// Makes the thread where this function is called to sleep for the specified value (in milliseconds).
  /// calls gsleep().
  static void sleep(GTime milliseconds);

  /// Prints an error message.
  /// calls gerror().
  /// By default, errors are printed on std::cerr. onError() changes how errors are managed.
  static void error(GString const& where, GString const& message, GObject const* = {});

  /// Changes the function that manages errors.
  /// - If _fun_ is not null, it will be called whenever gerror() or GApp::error() is called.
  /// - if _fun_ is null, errors are printed out on std::cerr (default behavior).
  /// @note if _object_ is not null this address is written in the error message.
  static void onError(std::function<void(GString const& where, GString const& message,
                                         GObject const* object)> fun);

  /// Changes the function that manages exceptions.
  /// - If _fun_ is not null, it will be called whenever an exception is thrown.
  /// - if _fun_ is null, a dialog box is opened to ask the user if she wants to quit or to continue.
  ///  This feature may allow saving important data before quiting the program, but there is no guarantee
  ///  it will work, especially when the exception was not thrown by the toolkit.
  static void onException(std::function<void(GString const& type, GString const& message,
                                             bool isGFatalError)> fun);

  /// Allows inspecting gadgets.
  /// A Meta-click or a Cmd-click (depending on the OS) on a gedget will open the inspector.
  static class GInspector& inspector();

  /// Shows gadget inspector.
  static void showInspector();
};

}
#endif
