//
// Graphics Layer
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Graphics_hpp
#define Guit_Graphics_hpp
#include <gcolor.hpp>
#include <gfont.hpp>
namespace guit {

class GFontValue;
class GIconValue;


/// Intersection of two rectangles.
class GClip {
public:
  /// empty clip
  GClip() : x1(0.f), x2(0.f), y1(0.f), y2(0.f) {}

  /// the clip is the intersection of the two rectangles.
  GClip(GRect const& rect1, GRect const& rect2);

  bool empty() const {return x2 <= x1 || y2 <= y1;}

  GRect frame() const {return GRect(x1, y1, x2-x1+1, y2-y1+1);}

  /// the clip is the intersection of the current rectangle with the rectangle given as an argument.
  void intersection(GRect const&);

  /// returns true is the clip contains this point.
  bool contain(GPoint const&) const;

  float x1, x2, y1, y2;
};



/// Class for drawing graphics in gagdgets.
/// Drawing must be performed in a callback function called on a GTrigger::paint trigger.
/// This callback function must create a Graphics object in the stack that won't be reused elsewhere:
/// @code
/// Canvas::Canvas() {
///    *this
///    << Background.white
///    << on::paint     / [this](auto& e) {paint(e);}
///    << on::resize    / [this](auto& e) {resized(e);}
///    << on::mousedown / [this](auto& e) {pressed(e);}
///    << on::mouseup   / [this](auto& e) {released(e);}
///    << on::mousedrag / [this](auto& e) {dragged(e);}
///    << on::mousemove / [this](auto& e) {moved(e);};
///  }
///
/// void Canvas::paint(Event& e) {
///   Graphics g(e);
///
///   for (auto& it : polylines) {
///     g.setColor(it->color);
///     g.setBgColor(it->bgcolor);
///     switch (it->mode) {
///       case PolylineMode:
///         g.drawPolyline(it->points);
///         break;
///       case PolygonMode:
///         g.drawPolygon(it->points);
///         break;
///     }
///   }
/// }
/// @endcode
/// All drawXxxx() methods draw relatively to frame(), which is the frame of the gadget
/// that received the paint event. The frame can be translated using translate().
///
/// OpenGL can be used for more advanced functionalities. Note that:
/// - glPos(), glX(), glY() return the position of the frame in OpenGL coordinates
///  (i.e. with the origin at the bottom-left corner of the window)
/// - the OpenGL state must be restored to ensure proper rendering (use glPushAttrib(),
///  glPopAttrib() and glPushMatrix(), glPopMatrix() accordingly).
///
class GGraphics {
protected:
  GGraphics(GWindow&);

public:
  GGraphics(GEvent&);

  /// position of the drawing area in window coordinates.
  /// origin = top-left corner of the window.
  /// @{
  float wx() const {return frame_.x;}
  float wy() const {return frame_.y;}
  GPoint wpos() const {return GPoint(frame_.x, frame_.y);}
  /// @}

  /// position of the drawing area in the window in GL coordinates
  /// origin = bottom-left corner of the window.
  /// @{
  float glx() const {return frame_.x;}
  float gly() const {return gly_;}
  GPoint glpos() const {return GPoint(frame_.x, gly_);}
  /// @}

  /// size of the drawing area.
  /// @{
  float width()  const {return frame_.w;}
  float height() const {return frame_.h;}
  GDim size() const {return GDim(frame_.w, frame_.h);}
  /// @}

  /// window where the graphics is painted.
  GWindow* window() {return win_;}

  /// frame of the drawing area in the window.
  /// origin = top-left corner of the window.
  GRect const& frame() const {return frame_;}

  /// translates the frame.
  void translate(float x, float y);
  
  void setScale(float scale);
  float scale() const {return scale_;}

  /// sets the clipping area in the window.
  /// the clip is in window coordinates.
  void setWinClip(GRect const&);
  void setWinClip(float x, float y, float w, float h);

  void save();
  void restore();
  void makeCurrent();
  void swapBuffers();
  void beginGL() {}
  void endGL() {}

  GRgba const& color() const {return color_;}
  void setColor(GRgba const* color);
  void setColor(GColor const* color);
  void setColor(GColor const& color);
  void setColor(GColor const& color, float alfa);
  void unsetColor();

  GRgba const& bgcolor() const {return bgcolor_;}
  void setBgColor(GRgba const* color);
  void setBgColor(GColor const* color);
  void setBgColor(GColor const& color);
  void setBgColor(GColor const& color, float alfa);
  void unsetBgColor();

  void setFont(GFont const& font);
  void setFont(GFontValue const& font);

  /// changes outline thickness.
  /// note: actual effect depends on GL implmenetaion
  void setThickness(float);

  /// draws a string.
  /// x, y = is the leftmost lower point. Returns the size of the rectangle enclosing the text.
  /// does not take into account multiline text
  /// @{
  GDim drawString(float x, float y, GString const& str);
  GDim drawString(GPoint const& p, GString const& str) {return drawString(p. x, p. y, str);}
  /// @}

  /// draws an image or an icon.
  /// @{
  void drawImage(GRect const&, GIcon&);
  void drawImage(float x, float y, float w, float h, GIconValue&);
  void drawImage(GRect const& r, GIconValue& ima) {drawImage(r.x, r.y, r.w, r.h, ima);}
  /// @}

  /// reads an image or an icon.
  /// @{
  static bool readImage(GIcon&);
  static bool readImage(GIconValue&);
  /// @}

  /// GL textures
  /// @{
  void drawTexture(float x, float y, float w, float h, GTexture texture, GRgba const* = {});
  static void deleteTexture(GTexture);
  /// @}

  void drawLine(float x1, float y1, float x2, float y2);
  void drawLine(GPoint const& p1, GPoint const& p2) {drawLine(p1.x,p1.y,p2.x,p2.y);}

  void drawPolyline(std::vector<GPoint> const& points);
  void drawPolygon(std::vector<GPoint> const& points);

  void drawRect(float x, float y, float w, float h);
  void drawRect(GRect const& r) {drawRect(r.x, r.y, r.w, r.h);}

  void drawRoundRect(float x, float y, float w, float h, float xradius, float yradius);
  void drawRoundRect(GRect const& r, float xradius, float yradius) {
    drawRoundRect(r.x, r.y, r.w, r.h, xradius, yradius);
  }

  void drawOval(float x, float y, float w, float h) {drawArc(x, y, w, h, 0,360,true);}
  void drawOval(GRect const& r) {drawArc(r.x, r.y, r.w, r.h, 0,360,true);}

  /// Draws an arc.
  /// - _rect_ is the enclosing rectangle, the center of the rectangle will be the center of the ellipse
  /// - _start_ if the starting angle in degrees (0 corresponds to  3 o'clock position),
  /// - _extent_ is the angular distance (in degrees) from _start_ to the ending angle.
  ///  A counter-clockwise rotation is performed if _extent_ is positive, a clockwise rotation otherwise.
  /// - if _pie_ is false drawArc() just draws an arc, otherwise it draws a "pie part", so as drawPie()
  /// @{
  void drawArc(float x, float y, float w, float h,
               float start, float extent, bool pie = false);
  void drawArc(GRect const& rect, float start, float extent, bool pie = false) {
    drawArc(rect.x,rect.y,rect.w,rect.h, start, extent, pie);
  }
  /// @}

  void drawPie(GRect const& rect, float start, float extent) {drawArc(rect,start,extent,true);}

  static bool pickOval(GRect const&, GPoint const&);
  static bool pickPie(GRect const&, float start, float extent, GPoint const&);

  /// @internal
  /// @{
public:
  bool drawfg_{}, drawbg_{};
  float gly_{}, thickness_{}, scale_{1.0f}, savex_{}, savey_{};
  GRgba color_, bgcolor_;
  GFontValue font_;
  GRect frame_;
  GWindow* win_;
  /// @}
};

}
#endif
