//
//  Dynamic loading of GUI specifications
//  Guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <sstream>
#include <gtrigger.hpp>
#include <gtextitem.hpp>
#include <gbox.hpp>
#include <gfile.hpp>
#include <gcolor.hpp>
#include <gscript.hpp>
#include <gobjectbase.hpp>
using namespace std;
namespace guit {

class GStringWrapper : public GObject {
public:
  GStringWrapper(GString const& str) : str(str) {}
  GString typeName() const override {return "/S";}
  GString str;
};

static GString* toString(GObject& obj)  {
  if (obj.typeName() != "/S") return nullptr;
  return &static_cast<GStringWrapper*>(&obj)->str;
}


class GTriggerWrapper : public GObject {
public:
  GTriggerWrapper(GTrigger* trigger) : trigger(trigger) {}
  GString typeName() const override {return "/T";}
  GTrigger* trigger;
};

static GTrigger* toTrigger(GObject& obj)  {
  if (obj.typeName() != "/T") return nullptr;
  return static_cast<GTriggerWrapper*>(&obj)->trigger;
}


static GFunction const* toFunction(GObject& obj) {
  if (obj.typeName() != "/F") return nullptr;
  return &static_cast<GFunctionWrapper*>(&obj)->fun;
}


using ObjPtr = gptr<GObject>;

class GScript {
public:
  GScript(std::istream&, GObjectBase*, GTextAttributes*, GString const* streamname);
  ObjPtr parseAll();
  ObjPtr parseExpr(ObjPtr leftobj, GString const& before, GString& end);
  ObjPtr parseLHS(ObjPtr leftobj, GString const& oper, GString& end);
  ObjPtr parseChildren(ObjPtr obj, GString& end);
  ObjPtr parseObject(ObjPtr leftobj, GString const& before, GString& end);
  bool parseArg(GString& arg);
  bool parseString(GString& arg);
  bool parseUntil(GString& arg, char end);
  bool parseConstant(GString& arg);
  ObjPtr parseFunBody(GString& end);
  ObjPtr createCond(GObject& left, GObject& right);
  std::istream& getNext(int& c);
  std::istream& getChar(int& c);
  void putBack(int c);
  char getToken(GString& tok);
  void addElem(GString const& var, GObject& obj, bool islocal);
  bool addChild(GObject& parent, GObject& child);
  bool bindTo(GObject& obj, GObject& other);
  void error(GString const& msg);

  int line_{}, level_{}, error_{};
  size_t cpos_{};
  std::istream* in_{};
  GString const* streamname_{};
  GObjectBase *globbase_{};
  GObjectBase locbase_;
  GTextAttributes* tokens_{};

};

static GPropList varattr_, stringattr_, propattr_, gadgetattr_;
static bool attrinit_;


GScript::GScript(std::istream& in, GObjectBase* base,
                 GTextAttributes* tokens, GString const* streamname) :
streamname_(streamname)
{
  GObjectBase::loadGadgetClasses();

  if (!attrinit_) {
    attrinit_ = true;
    varattr_.addBegin(*new GColor("navy"));
    propattr_.addBegin(*new GColor("darkgreen"));
    gadgetattr_.addBegin(*new GColor("steelblue"));
    //gadgetattr_.addBegin(*new GFont("bold"));
    stringattr_.addBegin(*new GColor("firebrick"));
  }

  in.imbue(std::locale("C"));
  in_ = &in;
  error_ = 0;
  line_ = 1;
  cpos_ = 0;
  globbase_ = base;
  tokens_ = tokens;
  if (tokens) tokens->clear();
}

Gadget* GLoader::loadFile(GString const& filename, GObjectBase* base,
                          GTextAttributes* tokens) {
  GIFile in(filename);
  if (!in) return nullptr;
  return load(in, base, tokens, filename);
}

Gadget* GLoader::loadBuffer(GString const& buffer, GObjectBase* base,
                            GTextAttributes* tokens) {
  std::istringstream in(buffer);
  return load(in, base, tokens, "buffer");
}


Gadget* GLoader::load(std::istream& in, GObjectBase* base,
                      GTextAttributes* tokens, GString const& streamname) {
  GScript s(in, base, tokens, &streamname);
  ObjPtr obj = s.parseAll();
  Gadget* g = obj ? obj->toGadget() : nullptr;
  if (g) g->omodes_.autodel = 0;
  return g;
}


void GScript::error(GString const& msg) {
  error_++;
  if (!streamname_) {
    //cerr << "GScript " << msg << " at line " << to_string(line_) << endl;
    gerror("GScript", msg + " at line " + to_string(line_));
  }
  else {
    gerror("GScript", msg + " at line " + to_string(line_)
           + " in "+ *streamname_);
  }
}


void GScript::addElem(GString const& var, GObject& obj, bool islocal) {
  GObjectBase::Status stat{};

  // non-local vars are put in locbase_ if there is no globbase_
  if (islocal || !globbase_) stat = locbase_.add(var, obj);
  else stat = globbase_->add(var, obj);

  if (stat == GObjectBase::InvalidName) {
    error("'"+ var+"' is not a valid identifier");
    return;
  }
  else if (stat == GObjectBase::NameExists) {
    error("'"+ var+"' already exists and can't be redefined");
    return;
  }
}


ObjPtr GScript::parseAll() {
  ObjPtr lastobj;

  while (in_) {
    GString var, tok, arg;

    // read var name or 'auto'
    if (!getToken(var) || var.empty()) {  // end of the file
      return lastobj;
    }

    bool islocal{};

    // read 'auto' if any
    if (var == "auto") {
      if (!getToken(var) || var != "&") {  // & needed after 'auto'
        error("expecting & after 'auto'");
        return nullptr;
      }
      else islocal = true;

      // read variable after auto
      if (!getToken(var) || var.empty()) {
        error("expecting variable");
        return nullptr;
      }
    }

    if (!getToken(tok) || (tok != "=" && tok != "<<")) {
      error("expecting = or <<");
      return nullptr;
    }

    GString end;
    ObjPtr obj;

    if (tok == "=") {
      obj = parseExpr(nullptr, "=", end);
      if (!obj) {
        error("invalid expression");
        return nullptr;
      }
      addElem(var, *obj, islocal);
    }
    else if (tok == "<<") {
       if ((obj = locbase_.getObject(var))) {   // search in local base
        //attr = &varattr_;
      }
      else if (globbase_ && (obj = globbase_->getObject(var))) {  // search in global base
       // attr = &varattr_;
      }
      if (!obj) {
        error("unknown variable '" + var +"'");
        return nullptr;
      }
      end = "<<";
      obj = parseChildren(obj, end);
      if (!obj) return nullptr;
    }

    lastobj = obj;

    if (end != ";") {
      error("expecting ;");
      return nullptr;
    }
  }

  return lastobj;
}


ObjPtr GScript::parseExpr(ObjPtr leftobj, GString const& before, GString& end) {
  end.clear();
  ObjPtr obj = parseLHS(leftobj, before, end);
  if (!obj) return nullptr;
  else return parseChildren(obj, end);
}


ObjPtr GScript::parseChildren(ObjPtr obj, GString& end) {
  if (!obj) return nullptr;

  while (true) {
    if (error_ != 0) return obj;  // errors were detected while parsing object

    if (end == "<<") {
      ObjPtr child = parseLHS(obj, "<<", end);
      if (!child || !addChild(*obj, *child)) return nullptr;
    }

    else if (end == "<<=") {
      ObjPtr right = parseLHS(obj, "<<=", end);
      if (!right || !bindTo(*obj, *right)) return nullptr;
    }

    else if (end == ")") {
      end = ")";
      return obj;
    }

    else if (end == ";") {
      if (level_ <= 0) {
        end = ";";
        return obj;
      }
      else {
        error("unexpected ;");
        return nullptr;
      }
    }

    else {
      error("syntax error");
      return nullptr;
    }
  }

  return obj;
}


ObjPtr GScript::parseLHS(ObjPtr leftobj, GString const& oper, GString& end) {
  int c = 0;
  getNext(c);

  if (c == '(') {
    GString tok;
    level_++;
    ObjPtr obj = parseExpr(leftobj, "(", tok);
    level_--;
    if (tok != ")") return nullptr;
    getToken(end);
    return obj;
  }

  putBack(c);

  ObjPtr first = parseObject(leftobj, oper, end);

  if (!first) return nullptr;
  if (end != "/") return first;

  // operator / was found
  while (true) {

    ObjPtr right = parseLHS(nullptr, "/", end);   // leftobj or fisrt?
    if (!right) {
      error("expecting expression after /");
      return nullptr;
    }

    first = createCond(*first, *right);
    if (!first) return nullptr;

    if (end != "/") return first;
  }
}


ObjPtr GScript::createCond(GObject& first, GObject& right) {

  GFunction const* func = toFunction(right);

  GProp* condprop{};
  if (!func) condprop = right.toProp();

  if (!func && !condprop) {
    error("prop or callback function expected after /");
    return nullptr;
  }

  auto* trigger = toTrigger(first);
  if (trigger) {
    if (func) return &(*trigger / *func);
    else return &(*trigger / condprop);
  }

  else if (auto* fprop = first.toProp()) {
    if (auto* c = fprop->toCond()) {
      return func ?  &(*c / *func) : &(*c / condprop);
    }
    else if (auto* b = fprop->toBoolExpr()) {
      return func ? &(*b / *func) : &(*b / condprop);
    }
  }

  error("invalid expression before /");
  return nullptr;
}


ObjPtr GScript::parseObject(ObjPtr leftobj, GString const& before, GString& end) {
  GString tok, arg;
  size_t cpos = cpos_, clen = 0;

  char tokstat = getToken(tok);
  if (!tokstat) return nullptr;

  if (tokstat == '[') return parseFunBody(end);

  if (!tok.empty() && !::isalpha(tok[0])) {
    error("invalid keyword '" + tok +"'");
    return nullptr;
  }

  char argstat = getToken(arg);
  if (!argstat || arg.empty()) {
    error("invalid argument '" + tok +"'");
    return nullptr;
  }

  if (tokstat == '(') {
    int endparent{};
    if (arg == ")") arg = ""; else getNext(endparent);
  }

  if (tokstat == '(' || tokstat == ':') getToken(end);
  else {
    end = arg;
    arg.clear();
  }

  ObjPtr obj;
  GPropList* attrprops{};

  if (tokstat == '"') {
    clen = +2;

    if (before == "=") {  // string
      attrprops = &stringattr_;
      obj = new GStringWrapper(tok);
    }

    else if (before == "<<=") {  // bind
      error("a string can't follow <<= operator:" +arg);
      return nullptr;
    }

    else {  // before can be << or (
      if (!leftobj) {
        error("syntax error after << operator:" +arg);
        return nullptr;
      }
      else if (leftobj->toBox()) {   // string added to box => convert to GItem
        attrprops = &gadgetattr_;
        obj = new GItem(tok);
      }
      else if (leftobj->toGadget()) {  // string added to nonbox => add string
        attrprops = &propattr_;
        auto* t = new GText(tok);
        t->setImmutable();
        obj = t;
      }
      else {
        error("syntax error after << operator:" +arg);
        return nullptr;
      }
    }
  }

  else if (tokstat == ':' && tok == "on") {   // trigger
    auto* t = GTrigger::getTrigger(arg);
    if (t) return new GTriggerWrapper(t);    // !!!mem
    else {
      error("unknown Trigger 'on::"+arg +"'");
      return nullptr;
    }
  }
  else {
    GString name;
    // class names can start with or without a G
    if (tok[0] == 'G') {
      name = tok.substr(1);
      clen++;
    }
    else name = tok;
    // create corresponding object; search variable if class does not exist

    if (auto* p = GPropType::createInstance(name, arg)) {   // if prop
      obj = p;
      attrprops = &propattr_;
    }
    else if (auto* g = GadgetType::createInstance(name, arg)) {  // if gadget
      obj = g;
      attrprops = &gadgetattr_;
    }
    else if ((obj = locbase_.getObject(tok))) {   // search in local base
      attrprops = &varattr_;
    }
    else if (globbase_ && (obj = globbase_->getObject(tok))) {  // search in global base
      attrprops = &varattr_;
    }
  }

  // add to list of retrievable tokens
  if (tokens_ && obj) {
    auto& a = *new GTextAttr(cpos, tok.size()+clen);
    if (attrprops) a.props_ = *attrprops;
    a.data_ = obj;
    //cerr <<"addtoken " << cpos << " " << tok.size()+clen << " " << tok << endl;
    tokens_->add(a);
  }

  if (obj) return obj;
  else {
    error("unknown class or variable '" + tok +"'");
    return nullptr;
  }
}


// [...]{var = truc;}   or    [...]{var << truc;}
ObjPtr GScript::parseFunBody(GString& end) {
  GString tok;

  getToken(tok);
  if (tok != "{") {
    error("function should start with { instead of '" + tok +"'");
    return nullptr;
  }

  getToken(tok);
  ObjPtr varobj;

  if (!(varobj = locbase_.getObject(tok))
      && (!globbase_ || !(varobj = globbase_->getObject(tok)))) {
    error("expecting variable followed by = instead of '" + tok +"'");
    return nullptr;
  }

  auto* v1 = varobj ? varobj->toProp() : nullptr;
  if (!v1) {
    error("variable '"+ tok + "' should be a Prop");
    return nullptr;
  }
  gptr<GVarProp> var = v1->toMutableProp();  // prevents autodelete
  if (!var) {
    error("variable '"+ tok + "' should be mutable");
    return nullptr;
  }

  getToken(tok);
  if (tok != "=") {
    error("expecting = instead of '" + tok +"'");
    return nullptr;
  }

  ObjPtr val = parseObject(nullptr, "=", tok);
  if (!val) {
    error("invalid value before '" + tok +"'");
    return nullptr;
  }

  if (tok != ";") {
    error("expecting ; instead of '" + tok +"'");
    return nullptr;
  }

  getToken(tok);
  if (tok != "}") {
    error("function should end with } instead of '" + tok +"'");
    return nullptr;
  }

  getToken(end);

  if (val) {
    if (auto* s = toString(*val)) {
      gptr<GText> txt = new GText(*s);   // prevent autodelete
      return new GFunctionWrapper( [var,txt]{var->set(*txt);} );
    }
    else if (auto* v = val->toProp()) {
      gptr<GProp> value = v;   // prevent autodelete
      return new GFunctionWrapper( [var,value]{var->set(*value);} );
    }
  }
  error("expecting a prop or a string before '" + tok +"'");
  return nullptr;
}


char GScript::getToken(GString& tok) {
  tok.clear();
  //char c = 0;
  int c = 0;

  while (getNext(c)) {
    //char c2 = 0;
    int c2 = 0;

    if (!::isalnum(c)) {
      switch (c) {
        case '"':
          return parseString(tok) ? '"': 0;

        case '(':
          return parseArg(tok) ? '(': 0;

        case '[':
          return parseUntil(tok, ']') ? '[' : 0;

        case '<':
          if ((c2 = in_->peek()) == '<') {
            getChar(c2);
            if ((c2 = in_->peek()) == '=') {
              getChar(c2);
              tok = "<<=";
            }
            else tok = "<<";
          }
          return '<';

        case '>':
          if ((c2 = in_->peek()) == '>') {
            getChar(c2);
            tok = ">>";
          }
          return '>';

        case '-':
          if ((c2 = in_->peek()) == '-') {
            getChar(c2);
            tok = "--";
          }
          return '-';

        default:
          tok = char(c);  // return operator
          return char(c);
      }
    }

    else {           // !isalnum(c) => token
      bool spacefound = false;
      tok = char(c);

      while (getChar(c)) {   // read token
        if (!spacefound && (::isalnum(c) || c == '_' || c == '.')) {
          tok += char(c);
        }
        else if (::isspace(c)) {
          spacefound = true;   // spaces between id and arg
        }
        else if (c == '(') {
          return '(';
        }
        else if (c == ':') {
           if ((c2 = in_->peek()) == ':') {
             getChar(c2);
             return ':'; // predef constant
           }
        }
        else {
          putBack(c);   // next field => putback
          return 1;
        }
      }

      if (in_->eof() && !tok.empty()) {
        putBack(c);
        return 1;
      }
    }
  }

  if (in_->eof()) {
    return 1;
  }

  if (in_->fail()) error("wrong token '"+ tok +"'");
  return 0;  // error or EOF
}


// precond: ( was already detected
bool GScript::parseArg(GString& arg) {
  arg.clear();
  int c = 0;
  while (getChar(c)) {
    if (c == '"') {     // string
      GString s;
      parseString(s);
      arg += s;
    }
    else if (c == ')') return true;  // end of argument
    else if (c == '\n') arg += ' ';
    else arg += char(c);
  }
  error("wrong argument '"+arg +"'");
  return false;
}


// precond: " was already detected
bool GScript::parseString(GString& arg) {
  bool stat = parseUntil(arg, '"');
  if (stat) {
    int c{};
    getChar(c);   // case "abcd"s
    if (c != 's') putBack(c);
  }
  return stat;
}

bool GScript::parseUntil(GString& arg, char end) {
  arg.clear();
  int c = 0;
  while (getChar(c)) {
    if (c == end) return true;  // end of sequence
    else arg += char(c);
  }
  error("unterminated "+ to_string(end) +" sequence '"+ arg +"'");
  return false;
}


// precond: :: was already detected
bool GScript::parseConstant(GString& arg) {
  arg.clear();
  int c = 0;
  while (getChar(c)) {
    if (::isalnum(c) || c == '_') arg += char(c);
    else {
      putBack(c);   // next field => putback
      return true;
    }
  }
  error("wrong constant");
  return false;
}

bool GScript::addChild(GObject& parent, GObject& child) {
  // cerr << "add: par:" <<  parent.typeName()
  // << " child: " << child.typeName()<< endl;

  if (auto* gchild = child.toGadget()) {
    if (auto* gpar = parent.toGadget()) {

      if (auto* boxpar = gpar->toBox()) {
        boxpar->add(*gchild);
        return true;
      }
      else if (auto* w = gchild->toWindow()) {
        gpar->add(*w);
        return true;
      }
      //else if (auto* m = gchild->toOverlay()) {
        else if (auto* m = gchild->toMenu()) {
        gpar->add(*m);
        return true;
      }
    }
    error("can't add gadget '" + child.typeName()
          + "' to '" + parent.typeName() +"'");
    return false;
  }

  else if (auto* pchild = child.toProp()) {
    if (auto* gpar = parent.toGadget()) {
      gpar->add(*pchild);
      return true;
    }
    else {
      error("can't add property " + child.typeName()
            + " to " + parent.typeName() + " (should be a Gadget)");
      return false;
    }
  }

  else if (auto* func = toFunction(child)) {
    if (auto* gpar = parent.toGadget()) {
      gpar->addFun(*func);                  // !!! memory
      return true;
    }
    else {
      error("can't add property " + child.typeName()
            + " to " + parent.typeName() + " (should be a Gadget)");
      return false;
    }
  }

  error("can't add " + child.typeName()
        + " to " + parent.typeName() + " (incompatible types)");
  return false;
}



bool GScript::bindTo(GObject& obj, GObject& other) {
  error("bind not yet implemented");
  return false;
}


void GScript::putBack(int c) {
  if (c == '\n') --line_;
  --cpos_;
  in_->putback(char(c));
}


std::istream& GScript::getChar(int& c) {
  if ((c = in_->get()) != EOF) {
    ++cpos_;
    if (c == '\n') ++line_;
  }
  return *in_;
}


std::istream& GScript::getNext(int& c) {
  //char c2{};
  int c2{};

  while (getChar(c)) {

    if (::isspace(c)) continue;

    // skip /* ... */ comments
    if (c == '/' && (c2 = in_->peek()) == '*') {
      getChar(c2);
      while (getChar(c)) {
        if (c == '*' && (c2 = in_->peek()) == '/') {
          getChar(c2);
          break;
        }
      }
      continue;
    }

    // skip // ... comments
    if (c == '/' && (c2 = in_->peek()) == '/') {
      getChar(c2);
      while (getChar(c)) {
        if (c == '\n') {
          if ((c2 = in_->peek()) == '\f') getChar(c2);
          break;
        }
      }
      continue;
    }

    if (!::isspace(c)) return *in_;
  }
  return *in_;
}

}
