//
// Strings
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gstring.hpp>
#include <regex>
#include <cstdio>
#include <gapp.hpp>
#include <gprefs.hpp>
using namespace std;
namespace guit {

/*
 #if defined(_MSC_VER)
 #include <BaseTsd.h>
 typedef SSIZE_T ssize_t;
 #endif
 */

void strtrim(GString& str, bool remquotes) {
  if (str.empty()) return;
  auto *p = str.data();
  auto *end = str.data() + str.size();

  // trim left
  while (p < end && ::isspace(*p)) ++p;
  if (p >= end) {str.clear(); return;}

  --end;
  // trim right
  while (end >= p && ::isspace(*end)) --end;
  if (end < p) {str.clear(); return;}

  if (remquotes) {
    if (*p == '"' || *p == '\'') ++p;
    if (*end == '"'  || *end == '\'') --end;
    if (end < p) {str.clear(); return;}
  }
  str.assign(p, size_t(end-p+1));
}

/*
GString strtrimmed(GString const& str, bool remquotes) {
  GString result(str);
  strtrim(result, remquotes);
  return result;
}

void unquote(GString& str) {
  trim(str, true);
}

GString unquoted(GString const& str) {
  GString result(str);
  trim(result, true);
  return result;
}*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

#define TRIM_(tok, mode) {\
  if (mode >= SplitMode::Trim) strtrim(tok, mode == SplitMode::RemQuotes);\
}

int strtoken(GString const& str, GString& token, size_t& from,
             GString const& delims, SplitMode mode) {
  bool spacesdel = delims.empty();
  size_t pos = from;
  char inquotes = 0;

  if (mode >= SplitMode::Trim) {   // skip spaces
    while (pos < str.size() && ::isspace(str[pos])) ++pos;
  }

  if (pos >= str.size()) {
    token.clear();
    from = pos;
    return -1;
  }

  // find delims
  while (pos < str.size()) {
      if (str[pos] == '"' && mode >= SplitMode::CheckQuotes) {
      if (inquotes == '"') inquotes = 0; else inquotes = '"';
    }
    else if (str[pos] == '\'' && mode >= SplitMode::CheckQuotes) {
      if (inquotes == '\'') inquotes = 0; else inquotes = '\'';
    }

    if (!inquotes) {
      if (spacesdel) {
        if (::isspace(str[pos])) break;
      }
      else {
        if (delims.find(str[pos]) != GString::npos) break;
      }
    }

    ++pos;
  }

  if (mode >= SplitMode::Trim) {   // skip spaces
    while (pos < str.size() && ::isspace(str[pos])) ++pos;
    if (pos < str.size()) {
      if (delims.find(str[pos]) == GString::npos) pos--;
    }
  }

  if (pos >= str.size()) {
    token.assign(str, from);
    TRIM_(token, mode);
    from = str.size();
    return 0;
  }
  else {
    token.assign(str, from, pos-from);
    TRIM_(token, mode);
    from = pos+1;
    return str[pos];
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

size_t strsplit(GString const& str, GString& left, GString& right,
                GString const& delims, SplitMode mode) {
  left.clear();
  right.clear();
  size_t pos{};

  int del = strtoken(str, left, pos, delims, mode);
  if (del == -1) return 0;

  if (pos >= str.size()) return 1;

  right.assign(str, pos);
  TRIM_(right, mode);
  return right.empty() ? 1 : 2;
}


size_t strsplit(GString const& str, std::vector<GString>& tokens,
                GString const& delims, int limit, SplitMode mode) {
  tokens.clear();
  size_t pos{};
  GString tok;

  if (limit <= 0) limit = 32767;

  while (int(tokens.size()) < limit && strtoken(str, tok, pos, delims, mode) != -1) {
    tokens.push_back(tok);
    /*
    if (adddelim && del > 0) {
      char buf[2] = {(char)del,0};
      tokens.push_back(buf);
    }*/
  }

  if (int(tokens.size()) < limit && pos < str.size()) {
    tok.assign(str, pos);
    //TRIM_(tok, mode);
    if (!tok.empty()) tokens.push_back(tok);
  }
  
  return tokens.size();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

size_t strregsplit(GString const& str, std::vector<GString>& tokens,
                   GString const& delim, int limit, bool trim) {
  bool spaces = delim.empty();

  tokens.clear();
  if (str.empty()) return 0;
  std::regex expr(spaces ? "[:space:]"+delim : delim);
  const char *p = str.c_str(), *prefix = p;

  auto trimRight = [&tokens](const char* beg, const char* end, bool atEnd) {
    while (end >= beg && ::isspace(*end)) --end;
    if (end < beg) {if (!atEnd) tokens.push_back("");}
    else tokens.push_back(GString(beg, size_t(end-beg+1)));
  };

  while (limit <= 0 || int(tokens.size()) < limit-1) {
    if (trim) { // trim left
      while (*p != 0 && ::isspace(*p)) ++p;
      if (*p == 0) {prefix = p; break;}
    }
    prefix = p;
    std::cmatch m;
    if (!std::regex_search(p, m, expr)) break;
    else {
      if (!trim) tokens.push_back(GString(prefix, size_t(p-prefix+m.position())));
      else (trimRight)(prefix, p+m.position()-1, false);
      p += m.position() + m.length();
      prefix = p;
    }
  }

  if (*prefix) {
    if (!trim) tokens.push_back(GString(prefix));
    else {
      while (*prefix != 0 && ::isspace(*prefix)) ++prefix;
      if (*prefix != 0) (trimRight)(prefix, prefix+::strlen(prefix)-1, true);
    }
  }
  return tokens.size();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

bool strblank(GString const& str) {
  if (str.empty()) return true;
  auto *p = str.c_str();
  while (*p != 0) {
    if (!::isspace(*p)) return false;
    ++p;
  }
  return true;
}

GString strlower(GString const& str) {
  GString s(str);
  for (auto& c : s) c = char(::tolower(c));
  return s;
}

GString strupper(GString const& str) {
  GString s(str);
  for (auto& c : s) c = char(::toupper(c));
  return s;
}

int strcompare(GString const& a, GString const& b, bool ignorecase) {
  if (!ignorecase) return a.compare(b);
  else {
    size_t len = a.length() < b.length() ? a.length() : b.length();
    for (size_t k=0; k < len; ++k) {
      if (::tolower(a[k]) < ::tolower(b[k])) return -1;
      else if (::tolower(a[k]) > ::tolower(b[k])) return +1;
    }
    return a.length() == b.length() ? 0 : a.length()<b.length() ? -1 : +1;
  }
}

bool strequal(GString const& a, GString const& b, bool ignorecase) {
  if (!ignorecase) return a == b;
  else if (a.length() != b.length()) return false;
  else return strcompare(a, b, true) == 0;
}

ssize_t strPosToIndex(GString const& text, ssize_t charpos, bool richtext) {
  if (charpos < 0) charpos = ssize_t(text.size());    // FAUX !!!!
  ssize_t bpos = 0, nextpos = 0;

  for (ssize_t k = 0; k < charpos; ++k) {
    nextpos = strNextIndex(text, bpos, richtext);
    if (nextpos == bpos) return bpos;  // end reached
    bpos = nextpos;
  }
  return bpos;
}

ssize_t strIndexToPos(GString const& text, ssize_t bytepos, bool richtext) {
  if (bytepos < 0) bytepos = ssize_t(text.size());
  ssize_t charpos = 0, bpos = 0, nextpos = 0;

  while (true) {
    if (bytepos <= bpos) return charpos;
    nextpos = strNextIndex(text, bpos, richtext);
    if (nextpos == bpos) return charpos;  // end reached
    bpos = nextpos;
    ++charpos;
  }
  return charpos;
}

// pos is a byte pos
size_t strNextIndex(GString const& text, ssize_t pos_, bool richtext) {
  if (pos_ < 0) return 0;
  size_t len = text.size(), pos = size_t(pos_);
  if (pos >= len) return len;

  if (richtext) {
    while (pos < len && text[pos] == '<') {
      ++pos;
      while (pos < len && text[pos] != '>') ++pos;
      if (pos < len && text[pos] == '>') ++pos;
      if (pos >= len) return len;
    }
  }
  unsigned char c = static_cast<unsigned char>(text[pos]);
  // 1 byte: val = 0xxxxxxx (val & 10... = 0)
  if ((c & 0x80) == 0) pos += 1;
  // 2 bytes: val = 110xxxxx (val & 1110... = 110...)
  else if ((c & 0xE0) == 0xC0) pos += 2;
  // 3 bytes: val = 1110xxxxx (val & 11110... = 1110...)
  else if ((c & 0xF0) == 0xE0) pos += 3;
  // 4 bytes
  else pos += 4;

  return pos > len ? len : pos;
}

// pos is a byte pos
size_t strPreviousIndex(GString const& text, ssize_t pos, bool richtext) {
  if (pos <= 0) return 0;
  if (size_t(pos) > text.size()) return text.size();

  if (richtext) {
    while (pos > 1 && text[size_t(pos-1)] == '>') {
      --pos;
      while (pos >= 0 && text[size_t(pos)] != '<') --pos;
      if (pos <= 0) return 0;
    }
  }
  if ((text[size_t(pos-1)] & 0x80) == 0) pos -= 1;
  else if (pos >= 2 && (text[size_t(pos-2)] & 0xE0) == 0xC0) pos -= 2;
  else if (pos >= 3 && (text[size_t(pos-3)] & 0xF0) == 0xE0) pos -= 3;
  else pos -= 4;
  return size_t(pos < 0 ? 0 : pos);
}

}
