//
// Main types & type conversions
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef Guit_Types_hpp
#define Guit_Types_hpp
#include <cstdint>
#include <memory>
#include <iostream>
#include <functional>
#include <string>
#include <vector>

#define GUIT_VERSION_MAJOR 0
#define GUIT_VERSION_MINOR 1
#define GUIT_VERSION_PATCH 1

#if defined(_WIN32)
#  undef min
#  undef max
#  undef small
using ssize_t = signed long;
#endif

namespace guit {

/// Alias for std::string.
using GString = std::string;

/// Vector of Strings
using GStrings = std::vector<GString>;

/// ID of a GClass or CSS Style.
using GId = uint16_t;

/// ID of an OpenGL texture (same type as GLuint).
using GTexture = uint32_t;

/// Time stamp.
using GTime = uint32_t;

/// Keyboard key code.
using GKeyCode = uint32_t;

// GEvent types (see gevent.hpp).
enum struct GEventType : uint32_t;

// GEvent Button and Modifier states (see gevent.hpp).
enum struct GEventState : uint32_t;

/// Location of handles for reshaping a Gadget.
enum struct GHandle : uint8_t;

/// Prototypes for callback functions and predicates.
/// @{
using GFunction = std::function<void()>;
using GEventFunction = std::function<void(class GEvent&)>;
using GPredicate = std::function<bool(class Gadget&)>;
/// @}


/// Prints an error message.
/// By default, errors are printed on std::cerr. GApp::onError() changes error management.
void gerror(GString const& where, GString const& message, class GObject const* = {});

/// Executes a function in the main thread.
/// _fun_ will be executed in the main thread, whatever the current thread.
/// @warning Functions that modify/update the GUI must always be called in the main thread.
void gexec(GFunction const& fun);

/// Makes the thread where this function is called to sleep for the specified value (in milliseconds).
void gsleep(GTime milliseconds);


/// X/Y Position.
/// @see GPos, which is a Gadget property.
struct GPoint {
  float x, y;

  GPoint() : x{}, y{} {}
  GPoint(float x, float y) : x(x), y(y) {}

  void set(float x_, float y_) {x = x_; y = y_;}
  GPoint operator+(GPoint const&) const;
  GPoint operator-(GPoint const&) const;
  bool operator==(GPoint const&) const;
  bool operator!=(GPoint const&) const;
  friend std::istream& operator>>(std::istream& in, GPoint&);
  friend std::ostream& operator<<(std::ostream& out, GPoint const&);
};


/// Width/Height Dimension.
/// @see GSize, which is a gadget property.
struct GDim {
  float w, h;

  GDim() : w{}, h{} {}
  GDim(float w, float h) : w(w), h(h) {}

  void set(float width, float height) {w = width; h = height;}
  bool operator==(const GDim&) const;
  bool operator!=(const GDim&) const;
  friend std::ostream& operator<<(std::ostream& out, GDim const&);
  friend std::istream& operator>>(std::istream& in, GDim&);
};


/// Rectangular area.
struct GRect {
  float x, y, w, h;

  GRect() : x{}, y{}, w{}, h{} {}
  GRect(float x, float y, float w, float h) : x(x), y(y), w(w), h(h) {}
  GRect(GPoint const& p, GDim const& d) : x(p.x), y(p.y), w(d.w), h(d.h) {}

  /// rectangle delimited by these two points.
  /// p1 does not have to be the left-top location nor p2 the right-bottom coordinate.
  GRect(GPoint const& p1, GPoint const& p2);

  GRect frame() const {return *this;}
  void set(float x, float y, float width, float height);
  void translate(float dx, float dy);
  bool contain(GPoint const& point) const {
    return point.x >= x && point.x <= x+w && point.y >= y && point.y <= y+h;
  }
  bool contain(GPoint const& point, float tolerance) const;
  bool intersect(GPoint const& line_point1, GPoint const& line_point2) const;
  bool operator==(const GRect&) const;
  bool operator!=(const GRect&) const;
  friend std::ostream& operator<<(std::ostream&, GRect const&);
  friend std::istream& operator>>(std::istream&, GRect&);
};


///  Vector of points.
struct GPoints : public std::vector<GPoint> {
  GRect frame() const;
  void translate(float dx, float dy);
  bool contain(GPoint const& point) const;
  bool contain(GRect const& rect) const;
  bool intersect(GRect const& rect) const;
  friend std::ostream& operator<<(std::ostream&, GPoints const&);
  friend std::istream& operator>>(std::istream&, GPoints&);
};


/// basic type conversions.
/// @{
inline bool gconvert(GString const& from, GString& to) {to = from; return true;}
bool gconvert(GString const& from, bool& to);
bool gconvert(GString const& from, float& to);
bool gconvert(GString const& from, int& to);
bool gconvert(GString const& from, int8_t& to);
bool gconvert(GString const& from, int16_t& to);

bool gconvert(const bool& from, GString& to);
bool gconvert(const float& from, GString& to);
inline bool gconvert(const int& from, GString& to) {to = std::to_string(from); return true;}
inline bool gconvert(const int8_t& from, GString& to) {to = std::to_string(from); return true;}
inline bool gconvert(const int16_t& from, GString& to) {to = std::to_string(from); return true;}

bool gconvert(const float& from, int& to);
inline bool gconvert(const int& from, float& to) {to = float(from); return true;}

inline bool gconvert(const float& from, float& to) {to = from; return true;}
inline bool gconvert(const int& from, int& to) {to = from; return true;}
inline bool gconvert(const bool& from, bool& to) {to = from; return true;}

bool gconvert(const double& from, float& to);
bool gconvert(const double& from, int& to);

template <typename T>
inline GString toString(T const& val) {GString s; gconvert(val,s); return s;}

template <typename T>
inline int toInt(T const& val) {int num{}; gconvert(val,num); return num;}

template <typename T>
inline float toFloat(T const& val) {float num{}; gconvert(val,num); return num;}
/// @}


// Gadget and prop metaclasses
class GadgetType;
class GPropType;

// Base class, superclass of Gadget and GProp
class GObject;

// Main Gadgets
class Gadget;
class GBox;
class GWindow;
class GShape;
class GItem;
class GTextItem;
class GValueItem;
class GChartBox;
class GChoiceBox;
class GScrollBox;
class GTable;
class GTRow;
class GTHead;
class GMenu;
class GMenuBar;

// Main Props
class GProp;
class GVarProp;
class GBool;
class GNumber;
class GInt;
class GFloat;
class GText;
class GIcon;
class GFont;
class GFormat;
class GColor;
class GBackground;
class GFlow;
class GFlex;
class GAlign;
class GGap;
class GPos;
class GSize;
class GMargin;
class GBorder;
class GChoice;
class GTextData;

// Timers and animations
class GTimer;
class GAnim;

// Conditional props
class GCond;
class GGuard;
class GOtherwise;

// Active Expressions
class GExpr;
class GTextExpr;
class GNumExpr;
class GBoolExpr;
class GExprBool;

// Statecharts
class GStateChart;
class GState;
class GHState;

// Callbacks and Events
class GTrigger;
class GEvent;
class GMouseEvent;
class GKeyEvent;
class GTouchEvent;
class GGestureEvent;
class GCaller;

// Styles, Themes and Graphics
class GTheme;
class GStyle;
class GRgba;
class GGraphics;
class GRender;
class GRenderGraphics;
class GTextAttributes;
class GTextAttr;

}
#endif
