//
//  Native SDL Layer
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_NativeSDL_hpp
#define Guit_NativeSDL_hpp
#include <functional>
#include <SDL.h>
#include <gevent.hpp>
#include <gcursor.hpp>
#ifdef __APPLE__
#include <objc/objc-runtime.h>
#endif

struct SDL_SysWMinfo;
struct SDL_Surface;

namespace guit {

class GNatFont;
class GFontResource;
class GLContext;
class GIconValue;

class GNatContext {
public:
  GNatContext(class GControlImpl&);
  ~GNatContext();

  void mainLoop();
  void subLoop(const bool& quit);

  void setMainLoopDelay(unsigned int ms_delay);
  void setSwapInterval(int interval);

  static uint32_t elapsedTime();
  GDim const& screenSize() const {return screensize_;}

  GPoint cursorPos() const;
  void moveCursor(GPoint const&);
  void setCursorShape(GCursor::Shape);
  void showCursor(bool state);
  void freezeCursor(bool state);
  void grabCursor(bool state);
  
  bool getClipboard(GString& text) const;
  bool setClipboard(GString const& text);
  
  bool startTimer(GTimer&, GTime delay);
  bool stopTimer(GTimer&);
  void pushUserEvent(GFunction* fun);
  
  static bool loadFont(GFontResource&);
  static void deleteFont(GFontResource&);
  static bool getFontInfo(GString const& path, GString& family, GString& style);
  static GTexture createFontTexture(GString const& text, GFontResource&);

  static GDim  textSize(GFontResource&, GString const& text);
  static float textWidth(GFontResource&, GString const& text);
  static float fontHeight(GFontResource&);
  static float fontAscent(GFontResource&);
  static float fontDescent(GFontResource&);

  static void readTexture(const char** xpm, GIconValue&);
  static void readTexture(GString const& filename, GIconValue&);
  static SDL_Surface* readSurface(const char** xpm);
  static SDL_Surface* readSurface(GString const& filename);
  static void deleteSurface(SDL_Surface*);

  /// @internal
  /// @{
  int  waitEvent(SDL_Event*);
  void processEvent(SDL_Event const&);
  void processWindowEvent(SDL_Event const&, GWindow&);
  /// @}

private:
  static GTexture surfaceToTexture(SDL_Surface*);  // deletes surf!
  class GControlImpl& ctrl_;
  Uint32 mainloop_delay_{10}, focus_winid_{}, focus_time_{};
  GDim screensize_;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GNatWindow {
public:
  GNatWindow(GWindow*, bool with_border, bool resizable);
  ~GNatWindow();
  
  Uint32 winID() const {return winid_;}
  void setClip(float x, float y, float w, float h);
  GLContext* glcontext() {return (GLContext*)glcontext_;}
  GLContext* createGLContext();
  void makeCurrent();
  void makeCurrent(GLContext*);
  void swapBuffers();
  
  void show(bool state);
  void paint();
  void raise();
  void minimize();
  void maximize();
  void restore();
  void move(float x, float y);
  void resize(float w, float h);

  bool setIcon(GString const& filename);
  void setTitle(GString const& title);
  bool setOpacity(float opacity);    ///< opacity is between 0.f and 1.f.
  void setResizable(bool resizable);
  void setMinSize(float w, float h);
  void setMaxSize(float w, float h);
  void setBordered(bool bordered);

  // MacOSX specific
#ifdef __APPLE__
  void stayOnTop();
  void stayOnTopOf(GNatWindow&);
  void joinAllSpaces();
  void setTitlebarTransparent(bool);
  void hideCloseButton(bool);
  void hideMiniaturizeButton(bool);
  void hideZoomButton(bool);
  void disallowFullScreen();
  id cocoaWindow();
#endif
  void createDelegate();
  void deleteDelegate();
  void resizeCB(uint32_t time, int w, int h);
  GWindow* window_{};
  SDL_Window* sdlwin_{};
  SDL_GLContext glcontext_{};
  Uint32 winid_{};
  bool usermove_{}, userresize_{};
#ifdef __APPLE__
  id delegate_{};
#endif
};

}
#endif
