//
// Wrappers for executing threads and calling callbacks without using a lambda
// (unstable, maybe partly removed)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Call_hpp
#define Guit_Call_hpp
#include <thread>
#include <gtypes.hpp>
namespace guit {

class GCaller {
public:
  virtual ~GCaller() = default;
  virtual void operator()(GEvent&) = 0;
  GCaller() = default;
  GCaller(GCaller const&) = delete;
  GCaller(GObject const&&) = delete;
  GCaller& operator=(GCaller const&) = delete;
  GCaller& operator=(GCaller const&&) = delete;
  /// @internal
  uint16_t refcount_{};
};


class GCallerF0 : public GCaller {
public:
  GCallerF0(GFunction const& fun) : fun_(fun) {}
  void operator()(GEvent&) override {fun_();}
  GFunction fun_;
};

class GCallerF1 : public GCaller {
public:
  GCallerF1(GEventFunction const& fun) : fun_(fun) {}
  void operator()(GEvent& e) override {fun_(e);}
  GEventFunction fun_;
};

template <class T>
class GCallerM0 : public GCaller {
public:
  GCallerM0(T* obj, void (T::*fun)()) : obj_(obj), fun_(fun) {}
  void operator()(GEvent&) override {(obj_->*fun_)();}
  T* obj_;
  void (T::*fun_)();
};

template <class T>
class GCallerM1 : public GCaller {
public:
  GCallerM1(T* obj, void (T::*fun)(GEvent&)) : obj_(obj), fun_(fun) {}
  void operator()(GEvent& e) override {(obj_->*fun_)(e);}
  T* obj_;
  void (T::*fun_)(GEvent&);
};


template <class T, class E>
class GCallerM2 : public GCaller {
public:
  GCallerM2(T* obj, void (T::*fun)(E&)) : obj_(obj), fun_(fun) {}
  void operator()(GEvent& e) override {
    if (auto* ee = dynamic_cast<E*>(&e)) (obj_->*fun_)(*ee);
    else gerror("GCaller","Invalid event type");
  }
  T* obj_;
  void (T::*fun_)(E&);
};


inline GCaller& Call(GFunction const& fun) {
  return *new GCallerF0(fun);
}

inline GCaller& Call(GEventFunction const& fun) {
  return *new GCallerF1(fun);
}

template <class T>
inline GCaller& Call(T* obj, void (T::*fun)()) {
  return *new GCallerM0<T>(obj, fun);
}

template <class T>
inline GCaller& Call(T* obj, void (T::*fun)(GEvent& e)) {
  return *new GCallerM1<T>(obj, fun);
}

template <class T, class E>
inline GCaller& Call(T* obj, void (T::*fun)(E& e)) {
  return *new GCallerM2<T,E>(obj, fun);
}



class GCallerThread : public GCaller {
public:
  std::thread thread_;
};

class GCallerThreadF0 : public GCallerThread {
public:
  GCallerThreadF0(GFunction const& fun) : fun_(fun) {}
  void operator()(GEvent&) override {fun_();}
  GFunction fun_;
};

template <class T>
class GCallerThreadM0 : public GCallerThread {
public:
  GCallerThreadM0(T* obj, void (T::*fun)()) : obj_(obj), fun_(fun) {}
  void operator()(GEvent&) override {(obj_->*fun_)();}
  T* obj_;
  void (T::*fun_)();
};


/// Can be added to a conditional expression to executes a callback function in a thread.
/// If there is a following action, it will be executed  AFTER the callback function has returned
/// and in the main thread. Hence, below, {result = data;} is performed once data is returned by doCalc()
/// and the GUI is changed in the main thread (which is mandatory to update it properly).
/// @code
/// auto& data = *new String();
/// auto& result = Label();
/// auto& button = Button("Do It")
/// << on::action / InThread([&]{doCalc(data);}) << [&]{result = data;};
/// @endcode
/// @see gcall() which allows executing a function in the main thread from another thread.
/// @{
inline GCallerThread& CallThread(GFunction const& fun) {
  return *new GCallerThreadF0(fun);
}

template <class T>
inline GCallerThread& CallThread(T* obj, void (T::*fun)()) {
  return *new GCallerThreadM0<T>(obj, fun);
}
/// @}

/// Can be added to a conditional expression to wait for a given amout of time.
/// delay is in ms. can be used to create simple animations.
/// @see GAnim fo how to create more complex animations.
inline GCallerThread& CallWait(GTime ms_delay) {
  return *new GCallerThreadF0([ms_delay]{gsleep(ms_delay);});
}

}
#endif
