//
// Prop that enforces exclusive selection in a Box
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gevent.hpp>
#include <gtrigger.hpp>
#include <gchoice.hpp>
#include <gbox.hpp>
#include <gitem.hpp>
#include <gtext.hpp>
#include <gcontrol.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP(Choice, GPropType::Choice)

GChoice& Choice(int index) {return *new GChoice(index);}
GChoice& Choice(int index, int min, int max) {return *new GChoice(index, min, max);}

GChoice::GChoice() {value_ = -1;}
GChoice::GChoice(int value) : GInt(value) {}
GChoice::GChoice(int value, int min, int max) : GInt(value, min, max) {}
GChoice::GChoice(GChoice const& from) : GInt(from.value_) {}
GChoice::GChoice(GString const& value) : GChoice() {set(value);}

GChoice* GChoice::clone(bool copy_value) const  {
  return copy_value ? new GChoice(*this) : new GChoice;
}

void GChoice::addModes(GadgetModes& m) {
  m.hasChoice = 1;
}

void GChoice::removeModes(GadgetModes& m) {
  m.hasChoice = 0;
}

void GChoice::increase(int const& val) {
  if (val > 0) {
    int maxch = 0;
    perform([&maxch](Gadget& g){
      if (auto* b = g.toBox()) {   // pas bon pour les RadioBoxes!
        auto chcount = b->childCount();
        if (maxch < chcount) maxch = chcount;
      }
    });
    if (value_ < maxch - val) set(value_ + val);
    else if (pmodes_.autoReset) set(min_);
  }
  else if (val < 0) {
    if (value_ >= min_ - val) set(value_ + val);
    else if (pmodes_.autoReset) set(max_);
  }
}

void GChoice::addBoxChild(GBox& boxref, Gadget* child, int pos) {
  if (!child) return;
  GBox* box = &boxref;

  if (box->choiceBehavior() == GBox::ChoiceBehavior::Select) {
    if (child->isActivable()) {
      child->modes_.checkable = 1; // make items automatically selectable
      if (pos == value()) child->modes_.checked = 1;  // select item added afterwards
    }
    if (child->modes_.checked) {
      // unselect child so that selectObject() will do something
      child->modes_.checked = 0;
      child->selectImpl(true, Gadget::FromChoice);
      if (auto* text = box->findProp<GText>()) *text = child->text();
    }
  }

  else if (box->choiceBehavior() == GBox::ChoiceBehavior::Show) {
    if (box->childCount() == 1) {
      child->show(true);
    }
    else if (pos == value()) {
      child->show(true);
    }
    else {
      child->show(false);
    }
  }
}

void GChoice::choose(GBox& box, Gadget* g) {
  int index = -1, k = 0;

  if (box.choiceBehavior() == GBox::ChoiceBehavior::Select) {
    for (auto& it : box.children()) {
      if (it->isCheckable()) {     // other gadgets are ignored
        if (it.get() == g) {index = k; break;}
        ++k;
      }
    }
  }
  else if (box.choiceBehavior() == GBox::ChoiceBehavior::Show) {
    for (auto& it : box.children()) {
      if (it.get() == g) {index = k; break;}
      else it->show(false);
      ++k;
    }
  }

  if (index != value() && (box.eventmask_ & GEventType::Changing) != 0) {
      GEvent e(GTrigger::changing, GCore::elapsedTime());
      e.aux_ = chosenItem(box);
      box.doFire(e);
    }

  set(index);

  // in case set(index) did nothing because the index was correct
  if (g && box.choiceBehavior() == GBox::ChoiceBehavior::Select) {
    g->modes_.checked = true;
  }
}


void GChoice::choose(GBox& box, GString const& name) {
  if (name.empty()) set(-1);

  if (isdigit(name[0])
      || (name[0] == '-' && name.size()>1 && isdigit(name[1]))
      || (name[0] == '+' && name.size()>1 && isdigit(name[1]))
      ) {
    set(atoi(name.data()));
    return;
  }

  // #1 => first item, etc.
  if (name.size() >= 2 && name[0] == '#' && std::isdigit(name[1])) {
    choose(std::atoi(name.data()+1));
    return;
  }
  int index = -1, k = 0;

  if (box.choiceBehavior() == GBox::ChoiceBehavior::Select) {
    for (auto& it : box.children()) {
      if (it->isCheckable()) {     // other gadgets are ignored
        if (it->toItem() && strequal(it->toItem()->text(), name, true)) {
          index = k;
          break;
        }
        //if (it->text().size() >= name.size()
        //    && it->text().compare(0, name.size(), name)==0)
        //{index = k; break;}
        ++k;
      }
    }
  }
  else if (box.choiceBehavior() == GBox::ChoiceBehavior::Show) {
    for (auto& it : box.children()) {
      if (it->toItem() && strequal(it->toItem()->text(), name, true)) {
        index = k;
        break;
      }
      //if (it->text().size() >= name.size()
      //    && it->text().compare(0, name.size(), name)==0)
      //{index = k; break;}
      ++k;
    }
  }
  set(index);
}


Gadget* GChoice::chosenItem(GBox const& box) const {
  if (value() < 0) return nullptr;
  int k = 0;

  if (box.choiceBehavior() == GBox::ChoiceBehavior::Select) {
    for (auto& it : box.children()) {
      if (it->isCheckable()) {     // other gadgets are ignored
        if (k == value()) return it;
        ++k;
      }
    }
  }
  else if (box.choiceBehavior() == GBox::ChoiceBehavior::Show) {
    for (auto& it : box.children()) {
      if (k == value()) return it;
      ++k;
    }
  }
  return nullptr;
}

static void notifyBox(GBox& box, Gadget* item) {
  GEvent e(on::changeaction, GCore::elapsedTime());
  e.aux_ = item;
  box.doFire(e);
}

void GChoice::radioCB(GBox& box) {
  bool found{};
  Gadget* tonotify{};
  int n = 0;

  if (box.choiceBehavior() == GBox::ChoiceBehavior::Show) {
    for (auto& it : box.children()) {
      if (!it) continue;
      if (n != value()) {
        it->show(false);
        //if (!(cond)(it, n)) it->show(false);
      }
      else {
        found = true;
        if (!it->isShown()) {
          it->show(true);
          tonotify = it.get();
          //notifyBox(box, it);
        }
      }
      n++;
    }
    if (!found) {
      set(0);  // first child displayed
      notifyBox(box, nullptr);
    }
    else if (tonotify) notifyBox(box, tonotify);   // see note below
  }

  else if (box.choiceBehavior() == GBox::ChoiceBehavior::Select) {
    for (auto& it : box.children()) {
      if (!it) continue;
      if (it->isCheckable()) {
        if (n != value()) {
          //if (!cond(it, n)) {
          it->selectImpl(false, Gadget::FromChoice);
        }
        else {
          found = true;
          if (!it->isChecked()) {
            it->selectImpl(true, Gadget::FromChoice);
            tonotify = it.get();
            //notifyBox(box, it); (see below)
          }
        }
        n++;  // take only into account selectable items
      }
    }
    // tonotify delayed because the callback might change the list, thus to
    // invalidate the iterator
    if (!found) notifyBox(box, nullptr);
    else if (tonotify) notifyBox(box, tonotify);
  }

  box.update();
}

}
