//
// Conditionals: Callbacks and conditional Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#include <thread>
#include <gapp.hpp>
#include <gcall.hpp>
#include <gcond.hpp>
#include <gevent.hpp>
#include <gbind.hpp>
#include <ganim.hpp>
#include <gbox.hpp>
#include <gitem.hpp>
#include <gtext.hpp>
#include <gexpr.hpp>
#include <gstyle.hpp>
#include <gcontrol.hpp>
using namespace std;
namespace guit {

GPropType& GCond::Type() {
  static auto& t(*new GPropType("Cond", GPropType::Cond));
  return t;
}

GString GCond::stringValue() const {return "<Cond>";}

GCond::~GCond() {
  if (!isnext_) trigger_.receiverDeleted();
}

bool GCond::onAdd(Gadget* g) {
  if (g && !isnext_) trigger_.receiverAdded(*g, *this);  // !!!!
  return true;
}

void GCond::onRemove(Gadget*) {
  if (!isnext_) trigger_.receiverRemoved();
}

void GCond::addNext(GCond* c) {
  GCond* last{};
  for (auto* it = this; it != nullptr; it = it->next_) last = it;
  last->next_ = c;
  c->isnext_ = true;
}

void GCond::activate(GEvent& e, bool state) {
  activate2(e, state);
  //GCore::paintAll();
}

/*
void GCond::paintOrNext(GEvent& e, bool state) {
  if (next_) {
    // NB: a major advantage of chaining callbacks is that the view is updated
    GCore::paintAll();
    next_->activate2(e, state);
  }
}*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GCondCall::GCondCall(GTrigger& t, GCaller& c)
: GCond(t), caller_(c) {
  ++caller_.refcount_;
}

GCondCall::~GCondCall() { 
  if (--caller_.refcount_ <= 0) {
    if (caller_.refcount_ == 0) {delete &caller_;}
    else gerror("~GCondCall","wrong reference count");
  }
}

GString GCondCall::stringValue() const {return "<Callback>";}

void GCondCall::fire(GEvent& e) {
  // ce cas ne doit arriver que si le precedent est un guard
  if (e.matches(trigger_,-1) > 0)
    activate(e, true);
}

void GCondCall::activate2(GEvent& e, bool state) {
  if (auto* c = dynamic_cast<GCallerThread*>(&caller_)) {
    if (state) {
      c->thread_ = std::thread(&GCondCall::threadCB, this);
      c->thread_.detach();
    }
    // else needed: else clause done by threadCB once this action is completed
    else {
      //paintOrNext(e, state);
      GCore::paintAll();
      if (next_) next_->activate2(e, state);
    }
  }
  else {      // NOT THREADED
    if (state) caller_(e);
    //paintOrNext(e, state);
    //GCore::paintAll();
    if (next_) next_->activate2(e, state);
  }
}

void GCondCall::threadCB() {
  //                   pas bon: plusieurs threads peuvent modifier la var  !!!!
  GEvent e(on::none, 0);
  caller_(e);
  GCore::paintAll();
  if (next_) gexec([this, &e]{next_->activate(e, true);});
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GGuard::GGuard(GTrigger& t, GBoolExpr* guard) :
GCond(t),
guard_(guard) {}

void GGuard::fire(GEvent& e) {
  e.hasguard_ = true;
  bool b = guard_ ? guard_->boolValue() : true;
  if (b && next_) next_->fire(e);
  if (b) e.guarddone_ = true;
}

void GGuard::activate2(GEvent& e, bool state) {
  e.hasguard_ = true;
  bool b = guard_ ? guard_->boolValue() : true;
  if (next_) next_->activate2(e, state);
  if (b) e.guarddone_ = true;
}

GOtherwise::GOtherwise(GTrigger& t, GBoolExpr* guard)
: GGuard(t, guard){}

void GOtherwise::fire(GEvent& e) {
  if (!e.guarddone_) {
    bool b = guard_ ? guard_->boolValue() : true;
    if (b && next_) next_->fire(e);
    if (b) e.guarddone_ = true;
  }
}

void GOtherwise::activate2(GEvent& e, bool state) {
  if (!e.guarddone_) {
    bool b = guard_ ? guard_->boolValue() : true;
    if (next_) next_->activate2(e, state);
    if (b) e.guarddone_ = true;
  }
}

// - - - - - Create a Guard

GGuard& GTrigger::operator[](GExprBool const& expr) {
  return *new GGuard(*this, new GBoolExpr(&expr.prop, GExpr::Identity,
                                          nullptr, GBoolExpr::TBool));
}

GGuard& Guard(GExprBool const& expr) {
  return *new GGuard(on::action, new GBoolExpr(&expr.prop, GExpr::Identity,
                                               nullptr, GBoolExpr::TBool));
}

GOtherwise& Otherwise() {
  return *new GOtherwise(on::action, nullptr);
}

GOtherwise& Otherwise(GExprBool const& expr) {
  return *new GOtherwise(on::action, new GBoolExpr(&expr.prop, GExpr::Identity,
                                                   nullptr, GBoolExpr::TBool));
}

/*
GGuard& GGuard::operator[](GExprBool const& expr) {
  return *new GGuard(on::action, new GBoolExpr(&expr.prop, GExpr::Identity,
                                               nullptr, GBoolExpr::TBool));
}

GOtherwise& GOtherwise::operator[](GExprBool const& expr) {
  return *new GOtherwise(on::action, new GBoolExpr(&expr.prop, GExpr::Identity,
                                                   nullptr, GBoolExpr::TBool));
}
 */

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GCondProp::GCondProp(GTrigger& t, GProp& p) : GCond(t),
active_(false),
child_(p) {
  if (auto* e = p.toExpr()) child_ = new GExprWrapper(*e);
}

GCondProp::GCondProp(GTrigger& t, Gadget& g) : GCond(t),
active_(false),
child_(g) {}

GCondProp::GCondProp(GTrigger& t, GString const& str) : GCond(t),
active_(false),
child_(new GText(str)) {}

GString GCondProp::stringValue() const {return "<CondProp>";}

GProp* GCondProp::prop() {
  return child_->toProp();
}

void GCondProp::apply(Gadget* g, GRender& r, Specif) {
  /* BUG!!!!
   * comme on ne fait pas addProp il n'y a pas de mise à jour auto !!!
   * voir aussi les anims ou c'est fait avec addProp(p, false)
   */
  /*
  if (g == xxx) {
     cerr << "\n\nAPPLY " << &trigger_ << endl;

    for (GCond* p = this; p ; p = p->next_){
      if (auto* c = dynamic_cast<GCondProp*>(p)) {
        cerr <<" cond: " << (c->child_ ? c->child_->typeName() :"Null")<< " / ";

        GProp* pp {};
        if (c->child_ && (pp = c->child_ ->toProp())) {
          if (GColor* ch1 = pp->to<GColor>()) cerr << ch1->value();
          else if (auto* ch2 = pp->to<GText>()) cerr << *ch2;

          cerr   << endl;
        }
      }
      else cerr << " callback" << endl;
    }
  }*/

  //if (active_ && child_) {
  if (child_) {
    /*
    if (g == xxx) {
      cerr << "\n $$$$$$$$$$ "
      << (active_ ? "YES" : "NO")
      << ": "
      << child_->typeName()
      << " ch " << child_->toProp()
      <<endl;
    }*/
    if (active_) {
      if (auto* p = child_->toProp()) {
        p->apply(g, r, GStyle::CondSpecif);
        //if (g) p->updateGadget(*g);
      }
    }
  }

  if (next_) next_->apply(g, r, GStyle::CondSpecif);
}

void GCondProp::fire(GEvent& e) {
  // ce cas ne doit arriver que si le precedent est un guard
  //if (isnext_) cerr << "GCondProp::fire isnext" << endl;
  
  // fait seulement sur le premier element de la liste !
  int val = e.matches(trigger_,-1);
  if (val != 0) {      // doesn't match if val is 0
    if (val > 0) activate(e, true);
    else activate(e, false);
  }
}

void GCondProp::activate2(GEvent& e, bool state) {

  if (child_) {
    //cerr << "activate " << (child_->typeName()) << endl;
    active_ = state;
    Gadget* g{};

    if (auto* p = child_->toProp()) {
      if (auto* t = trigger_.toExprTrigger()) {g = t->receiver();}
      else g = e.gadget_;
      /*
      if (g == xxx) {
        cerr << "\n\activate " << state << " " << &trigger_
        << " " << p->typeName()
        << endl;
      }*/

      if (g) {
        if (state) {
          if (auto* a = p->toAnim()) a->start();
          g->addProp(*p, false);
        }
        else {
          if (auto* a = p->toAnim()) a->stop();
          g->removeProp(*p, false);
        }
      }
    }
    else if ((g = child_->toGadget())) {
      g->show(state);
    }
  }

  if (next_) next_->activate2(e, state);
  //else GCore::paintAll();
}


bool GCondProp::isEquivalent(GProp const& prev) const {
  if (!child_) return false;  // pas ideal mais le plus coherent

  // replace prev only if it is a condprop (if prop is un-cond it must remain)
  GCondProp* prev_cond{};
  if (prev.isCondProp()) prev_cond = static_cast<GCondProp*>(const_cast<GProp*>(&prev));
  if (!prev_cond) return false;
  if (!prev_cond->child_) return true;
  
  // can replace if OK both for trigger and prop
  auto* p = child_->toProp();
  auto* prevp = prev_cond->prop();
  if (!p || !prevp || !p->isEquivalent(*prevp)) return false;

  return trigger_ == prev_cond->trigger_;
}


bool GCondProp::onAdd(Gadget* par) {
  if (!child_) return false;   // dont add null prop
  if (!par) return true;
  GProp* p = child_->toProp();

  // if Text in a box don't add multiple times (especially if in fact a GString)
  if (p) {
    if (auto* txt = p->toText()) {
      if (auto* box = par->toBox()) {
        child_ = *new GLabel << txt;
      }
    }
  }

  // if gadget, adds it, it will shown/hidden
  if (auto* g = child_->toGadget()) {
    if (auto* box = par->toBox()) {
      g->show(false);
      box->add(*g);
    }
    else gerror("GCondProp","A gadget can only be added to a box");
  }

  trigger_.receiverAdded(*par, *this);

  if (trigger_.toExprTrigger()) return true;
  
  if (trigger_.isEventTrigger()) {
    // if on::change / check / uncheck the gadget becomes checkable (and checked if needed)
    if ((trigger_.eventMask() & GEventType::Change) != 0) par->modes_.checkable = 1;

    if (p) {
      if (!trigger_.filter() && !par->isChecked()) {
        active_ = true;
        p->onAdd(par);
      }
      else if (trigger_.filter() && par->isChecked()) {
        active_ = true;
        p->onAdd(par);
      }
    }
  }
  return true;
}


// - - - Create a GCond calling a function (or add a function to a GCond)

GCond& operator/(GTrigger& t, GFunction const& fun) {
  return *new GCondCall(t, *new GCallerF0(fun));
}

GCond& operator/(GTrigger& t, GEventFunction const& fun) {
  return *new GCondCall(t, *new GCallerF1(fun));
}

GCond& operator/(GTrigger& t, GCaller& c) {
  return *new GCondCall(t, c);
}


GCond& operator/(GBoolExpr& e, GFunction const& fun) {
  return *new GCondCall(*new GExprTrigger(e), *new GCallerF0(fun));
}

GCond& operator/(GBoolExpr& e, GEventFunction const& fun) {
  return *new GCondCall(*new GExprTrigger(e), *new GCallerF1(fun));
}

GCond& operator/(GBoolExpr& e, GCaller& c) {
  return *new GCondCall(*new GExprTrigger(e), c);
}


GCond& operator/(GCond& c, GFunction const& fun) {
  // !! ATT                                      !!! la prop n'a pas de parent !!
  c.addNext(new GCondCall(c.trigger(), *new GCallerF0(fun)));
  return c;
}

GCond& operator/(GCond& c, GEventFunction const& fun) {
  c.addNext(new GCondCall(c.trigger(), *new GCallerF1(fun)));
  return c;
}

GCond& operator/(GCond& c, GCaller& call) {
  c.addNext(new GCondCall(c.trigger(), call));
  return c;
}

// - - - - - Create a GCond adding a prop (or add an additional prop)

GCond& operator/(GTrigger& t, GString const& str) {
  return *new GCondProp(t, str);
}

GCond& operator/(GTrigger& t, GProp& prop) {
  return *new GCondProp(t, prop);
}

GCond& operator/(GTrigger& t, GProp* prop) {
  if (prop) return *new GCondProp(t, *prop);
  else {
    gerror("operator/", "null prop");
    return *new GCond(on::none);
  }
}

GCond& operator/(GTrigger& t, Gadget& g) {
  return *new GCondProp(t, g);
}

GCond& operator/(GTrigger& t, Gadget* g) {
  if (g) return *new GCondProp(t, *g);
  else {
    gerror("operator/", "null gadget");
    return *new GCond(on::none);
  }
}


GCond& operator/(GBoolExpr& e, GString const& str) {
  return *new GCondProp(*new GExprTrigger(e), str);
}

GCond& operator/(GBoolExpr& e, GProp& prop) {
  return *new GCondProp(*new GExprTrigger(e), prop);
}

GCond& operator/(GBoolExpr& e, GProp* prop) {
  if (prop) return *new GCondProp(*new GExprTrigger(e), *prop);
  else {
    gerror("operator/", "null child argument");
    return *new GCond(GTrigger::none);
  }
}

GCond& operator/(GBoolExpr& e, Gadget& g) {
  return *new GCondProp(*new GExprTrigger(e), g);
}

GCond& operator/(GBoolExpr& e, Gadget* g) {
  if (g) return *new GCondProp(*new GExprTrigger(e), *g);
  else {
    gerror("operator/", "null gadget");
    return *new GCond(on::none);
  }
}


GCond& operator/(GCond& c, GString const& str) {
  c.addNext(new GCondProp(c.trigger(), str));
  return c;
}

GCond& operator/(GCond& c, GProp& prop) {
  c.addNext(new GCondProp(c.trigger(), prop));
  return c;
}

GCond& operator/(GCond& c, GProp* prop) {
  if (prop) c.addNext(new GCondProp(c.trigger(), *prop));
  else gerror("operator/", "argument can't be a null prop");
  return c;
}

}
