//
// Conditionals: Callbacks and conditional Props and expressions
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitCond_hpp
#define GuitCond_hpp
#include <gprop.hpp>
#include <gptr.hpp>
namespace guit {

/// Conditional Expression.
/// A Conditional Expressions has the following form:
/// @code
///    action
///    trigger / action
///    trigger / action1 / action2 / etc.
///    trigger / transition / action1 /action2 / etc.
/// @endcode
///
/// The *trigger* specifies when the *action* is executed. Several actions can be specified,
/// and they will be executed in this order. Optionally, there can be a state *transition*
/// between the trigger and the first action.
///
/// A *trigger* can be (see details and examples below):
/// - Omitted (only for callbacks), the on::action default trigger is then used
/// - An Event Type such as on::mousepress
/// - An boolean Active Expression (see GExpr) that can contain props, states and timers
///
/// An *action* can be:
/// - A Callback function or lambda
/// - A Prop or an Active Expression, which is activated depending on the trigger's state.
///
/// A *transition* is a state transition such as: (state1 >> state2). See GState.
///
/// @section Callback functions
/// @code
///   auto& btn = Button("Do It")
///   << [this]{doThis();}
///   << on::action / [this]{doThat();}
///   << on::mousedown / [this]{doDown();}
///   << on::mouseup   / [this]{doUp();};
/// @endcode
///
/// The left part is the trigger, the right part is a lambda that is executed when the trigger is fired.
/// The trigger can be omitted, in which case the default on::action trigger is used (see below).
/// The trigger can also be a boolean Active Expression, as explained in a later section.
///
/// Supposing that doIt(), doDown() and doUp() are methods:
/// - this->doThis() and this->doThat() will be called when the button is clicked
/// - this->doDown() when the button is pressed
/// - this->doUp() when the button is released
///
/// The meaning of the default trigger depends on the gadget it is added to, e.g. it is equivalent to:
/// - on::click for a stateless button/item (see GItem)
/// - on::change for a state button/item
/// - on::checked for a state button/item that is in an exclusive box (see ChoiceGadget).
///
/// Lambdas capture *this*, as in this example. They can also capture variables by value (using [=])
/// or by reference (using [&]). C++ references must be captured by reference, and pointers by value.
///
/// The lambdas can have an optional GEvent& argument:
/// @code
///   auto& btn = HBox()
///   << on::paint     / [this](auto& e) {doPaint(e);}
///   << on::mousedown / [this](auto& e) {doDown(asMouseEvent(e));}
///   << on::mouseup   / [this](auto& e) {doUp(asMouseEvent(e));};
///   << on::mousedrag / [this](auto& e) {doDrag(asMouseEvent(e));};
/// @endcode
/// The GEvent argument is the event that fired the trigger. In this example, this alllows doPaint()
/// to draw graphics and doDown() etc. to retreive the position of the mouse
///
/// @section Conditional Properties
/// @code
///     auto& btn = Button
///     << on::checked   / "selected" / Color("red")
///     << on::unchecked / "idle" / Color("blue");
/// @endcode
/// The left part is the trigger, the right part is a Prop that is activated when the trigger is fired.
/// Thus, the button will display "selected" in red color when it is checked, and "idle" in blue color
///  when it is unchecked.
///
/// When the trigger has a counterpart (e.g. on::checked vs. on::unchecked), the Prop is deactivated
/// when the counterpart is fired. Thus, the previous example is equivalent to:
/// /// @code
///     auto& btn = Button
///     << "idle" / GColor::blue;
///     << on::checked / "selected" / GColor::red;
/// @endcode
///
/// Note that attributed text could also have been used:
/// @code
///     auto& btn = Button
///     << on::checked   / "<color=red>selected"
///     << on::unchecked / "<color=blue>idle";
/// @endcode
///
/// @section Conditional Active Expressions
/// @code
///     auto& euros = Float(100);   // 100 Euros
///     auto& rate = Float(1.2);    // Euro => Dollar conversion rate
///
///     auto& btn = Button
///     << on::checked   / "The value is " + (~euros * ~rate) + "$"
///     << on::unchecked / "The value is " + ~euros + "€"
/// @endcode
/// The right part is an Active Expression that is evaluated dynamically. It is automatically updated
/// when the value of one of its variables (which must be preceeded by a ~) is changed.
/// Active Expressions are activated/deactivated in the same way as Props.
///
/// @section Boolean Triggers (Boolean Active Expressions)
/// The trigger can also be a boolean Active Expression, which is reevaluated when the value of one
/// of its variables is changed:
/// - if the action is a callback, it is executed whenever the value of the expression becomes true
/// - if the action is a conditional prop or expression, it is activated when the value becomes true
///  and deactivated when the value becomes false.
/// @code
///     auto& mylabel = Label
///     (~myint == 1)         / GColor::red  // when myint becomes equal to 1
///     (~mytext == "danger") / CGColor:red  // when myint becomes equal to "danger"
///     (~mystate == true)    / GColor::red  // when mystate is entered
///     + mystate             / GColor::red  // when mystate is entered
///     - mystate             / GColor::red  // when mystate is exxited
///     (~mytimer > 5)        / GColor::red  // when mytimer has been fired 5 times
///     (~a + ~b == ~c)       / GColor::red  // when a + b == c
/// @endcode
///
/// @section Guards and State Transitions
/// Guards and State Transitions can also be part of conditional expressions: see GGuard, on::guard,
/// and GState, GHState, GStateChart.
///
/// @see GExpr for examples and more info on Active Expressions.
class GCond : public GProp {
public:
  GCond(GTrigger& t) : trigger_(t) {}
  ~GCond();

  bool isEquivalent(const GProp&) const override {return false;}

  static GPropType& Type();
  GPropType& type() const override {return Type();}
  
  /// @internal
  /// @{
  GCond* clone(bool copy_value) const override {return nullptr;}  ///< can't be cloned.
  GCond* toCond() override {return this;}
  GString stringValue() const override;
  GTrigger& trigger() const {return trigger_;}
  void activate(GEvent&, bool state);
  virtual void activate2(GEvent&, bool state) {};
  void fire(GEvent&) override {}
  void addNext(GCond*);
  GCond* next() {return next_;}
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  
  GTrigger& trigger_;
  gptr<GCond> next_;
  bool isnext_{};
  /// @}
};


/// Guard for conditional statements.
class GGuard : public GCond {
public:
  /// Adds a Guard to a to a trigger.
  /// The Guard is a boolean active expression, see Expr.
  GGuard& operator[](GExprBool const& expr);

  /// @internal
  /// @{
  GGuard(GTrigger& t, GBoolExpr* guard);
  GString stringValue() const override {return "<Guard>";}
  void activate2(GEvent&, bool state) override;
  void fire(GEvent&) override;
  gptr<GBoolExpr> guard_;
  /// @}
};

/// Returns a guard (see GGuard).
GGuard& Guard(GExprBool const& expr);


/// Otherwise guard.
class GOtherwise : public GGuard {
public:
  /// Adds an Otherwise Guard to a to a trigger.
  /// The Guard is a boolean active expression, see Expr.
  GOtherwise& operator[](GExprBool const& expr);

  /// @internal
  /// @{
  GOtherwise(GTrigger& t, GBoolExpr* guard);
  GString stringValue() const override {return "<Otherwise>";}
  void activate2(GEvent&, bool state) override;
  void fire(GEvent&) override;
  /// @}
};

/// Returns an Otherwise guard (see GOtherwise).
/// @{
GOtherwise& Otherwise();
GOtherwise& Otherwise(GExprBool const& expr);
/// @}


/// Returns a conditional statement that calls a callback function; see GCond.
/// @{
GCond& operator/(GTrigger& trigger, GFunction const& fun);
GCond& operator/(GTrigger& trigger, GEventFunction const& fun);
GCond& operator/(GTrigger& trigger, GCaller&);

GCond& operator/(GBoolExpr& expr, GFunction const& fun);
GCond& operator/(GBoolExpr& expr, GEventFunction const& fun);
GCond& operator/(GBoolExpr& expr, GCaller&);

GCond& operator/(GCond&, GFunction const& fun);
GCond& operator/(GCond&, GEventFunction const& fun);
GCond& operator/(GCond&, GCaller& fun);
/// @}

/// Returns a conditional statement that activates a Prop or an Active Expression; see GCond.
/// @{
GCond& operator/(GTrigger& trigger, GString const& str);
GCond& operator/(GTrigger& trigger, GProp& prop);
GCond& operator/(GTrigger& trigger, GProp* prop);
GCond& operator/(GTrigger& trigger, Gadget&);
GCond& operator/(GTrigger& trigger, Gadget*);

GCond& operator/(GBoolExpr& expr, GString const& str);
GCond& operator/(GBoolExpr& expr, GProp& prop);
GCond& operator/(GBoolExpr& expr, GProp* prop);
GCond& operator/(GBoolExpr& expr, Gadget&);
GCond& operator/(GBoolExpr& expr, Gadget*);

GCond& operator/(GCond&, GString const& str);
GCond& operator/(GCond&, GProp& prop);
GCond& operator/(GCond&, GProp* prop);
/// @}


/// @internal.
class GCondCall : public GCond {
public:
  GCondCall(GTrigger&, GCaller&);
  ~GCondCall();
  GString stringValue() const override;

  /// @internal
  /// @{
  void activate2(GEvent&, bool state) override;
  void fire(GEvent&) override;
  void threadCB();
  GCaller& caller_;
  /// @}
};


/// @internal.
class GCondProp : public GCond {
public:
  GCondProp(GTrigger&, GProp&);
  GCondProp(GTrigger&, Gadget&);
  GCondProp(GTrigger&, GString const&);
  bool isCondProp() const override {return true;}
  GProp* prop();
  GString stringValue() const override;

  /// @internal
  /// @{
  void fire(GEvent&) override;
  void activate2(GEvent&, bool state) override;
  bool isEquivalent(const GProp& other) const override;
  bool onAdd(Gadget*) override;
  void apply(Gadget*, GRender&, Specif) override;
  bool active_;
  gptr<GObject> child_;
  /// @}
};


}
#endif
