//
// Prop that specifies the orientation of a gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitFlow_hpp
#define GuitFlow_hpp
#include <gvarprop.hpp>
namespace guit {

class GFlowValue {
public:
  GFlowValue() = default;
  GFlowValue(int8_t kind, int16_t gridsize) : kind(kind), gridsize(gridsize) {}
  bool operator==(GFlowValue const&) const;
  bool operator!=(GFlowValue const&) const;
  int8_t kind{};
  int16_t gridsize{};
};

inline bool gconvert(GFlowValue const& from, GFlowValue& to) {to = from; return true;}
bool gconvert(GString const& from, GFlowValue& to);
bool gconvert(GFlowValue const& from, GString& to);


/// Specifies whether a box is laid out vertically, horizontally, as a grid, a pie or a stack.
/// This prop specifies the main direction of a GBox (and subclasses).
/// When applicable, the cross direction is the other direction, e.g. if the main direction is _row_
/// (horizontal) then the cross direction is _column_ (vertical).
///
/// Note that, whatever the value of this prop:
/// - Non-box gadgets are always horizontal (_row_ direction)
/// - GTable gadgets are always vertical (_column_ direction)
/// - GTRow gadgets (Table Rows) are always horizontal  (_row_ direction).
///
class GFlow : public GVarProp_<GFlowValue> {
public:
  using Super = GVarProp_<GFlowValue>;
  enum Kind : int8_t {Column, Row, Pie, Stack};

  /// column means vertical , row  horizontal, pie circular layout.
  static GFlow column, row, pie, stack;

  /// Constructor.
  /// - the default direction of the no-arg constr is _row_ (horizontal)
  /// - _gridsize_ is the number of columns in a grid. It takes effect only if type is GFlow::Row.
  /// @{
  GFlow();
  GFlow(GString const& value);
  GFlow(Kind kind, int16_t gridsize = 0);
  GFlow(GFlow const&);
  GFlow(GFlowValue const&);
  /// }

  GFlow* clone(bool copy_value) const override;

  GFlow& operator=(GFlow const& flow) {set(flow); return *this;}
  GFlow& operator=(GString const& flow) {set(flow); return *this;}

  int16_t gridSize() const {return value_.gridsize;}

  using GType = GPropType_<GFlow,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  bool onAdd(Gadget*) override;
  void updateGadget(Gadget&) override;
  void addModes(GadgetModes&) override;
  void apply(Gadget* g, GRender& r, Specif) final;
  /// @}

private:
  friend class Gadget;
};

/// Returns a GFlow created in the heap.
/// @{
inline GFlow& Flow() {return *new GFlow();}
inline GFlow& Flow(GString const& value) {return *new GFlow(value);}
/// @}

}
#endif
