//
// Font Prop
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <unordered_map>
#include <gfont.hpp>
#include <gfontbase.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gtextrender.hpp>
#include <gnative.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

/*
 class GFontInfo {
 public:
 //char style1{}, weight{};
 GString family, style;
 };*/


GIMPLEMENT_PROP_TYPE(Font, GPropType::Font)

static GCore::BeginPlainObjects begin;
GFont
GFont::regular("12px sans-serif", 0),
GFont::bold("bold 12px sans-serif", 1),
GFont::italic("italic 12px sans-serif", 2),
GFont::large("16px sans-serif", 3),
GFont::small("11px sans-serif", 4);
static GCore::EndPlainObjects end;

GFontValue::GFontValue() {
  GFontResource::set(impl, GFontResource::emptyFont());
}

GFontValue::~GFontValue() {
  GFontResource::set(impl, nullptr);
}

GFontValue::GFontValue(GString const& value) {
  gconvert(value, *this);
}

GFontValue::GFontValue(GFontValue const& from) {
  GFontResource::set(impl, from.impl);
}

GFontValue& GFontValue::operator=(GFontValue const& from) {
  GFontResource::set(impl, from.impl);
  return *this;
}

bool GFontValue::operator==(const GFontValue& other) const {
  return impl == other.impl;
}

bool GFontValue::operator!=(const GFontValue& other) const {
  return impl != other.impl;
}

uint8_t GFontValue::size() const {
  if (!realize()) return 0;
  return impl->size();
}

float GFontValue::fontHeight() const {
  if (!realize()) return 0.f;
  return GNatContext::fontHeight(*impl);
}

float GFontValue::fontAscent() const {
  if (!realize()) return 0.f;
  return GNatContext::fontAscent(*impl);
}

float GFontValue::fontDescent() const {
  if (!realize()) return 0.f;
  return GNatContext::fontDescent(*impl);
}

float GFontValue::fontWidth() const {
  return GNatContext::textSize(*impl, "G").w;
}

GDim GFontValue::textSize(GString const& str, bool richtext,
                          GTextAttributes const* attrs) const {
  if (!realize()) return GDim();
  if (!richtext) return GNatContext::textSize(*impl, str);
  else return GTextRender(*this, richtext, nullptr, 0, 0, attrs).getSize(str);
}

bool GFontValue::realize() const {
  if (!impl) throw GFatalError("GFontValue::realize","Unspecified font");
  if (impl->realize()) return true;
  impl = GFont::regular.value().impl;
  return impl->natfont();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static void parseFont(GString const& fspec, GFontValue& to) {
  GString style = "regular";
  int size = 12;
  GString families = "sans-serif";

  GString token, previous;
  size_t pos = 0, familypos = 0;
  int count = 0;

  while (strtoken(fspec, token, pos) != -1) {
    if (::isdigit(token[0])) {
      size = std::atoi(token.data());
      familypos = pos;

      if (count == 1) style = previous;
      else if (count > 1) gerror("GFont","invalid font specification: "+fspec);
      break;
    }
    count++;
    previous = token;
  }

  if (count > 0 && familypos < fspec.size() && !::isdigit(fspec[familypos])) {
    families = fspec.substr(familypos);
    strtrim(families);
  }
  /*
  cerr
  << "style '"<< style <<"'\n"
  << "size '"<< size <<"'\n"
  << "families '"<< families <<"'\n"
  <<endl;
   */
  GFontResource::set(to.impl, GFontResource::obtainFont(style, uint8_t(size), families));
}

bool gconvert(GString const& from, GFontValue& to) {
  GFontResource::set(to.impl, GFontResource::emptyFont());
  if (from.empty()) return false;

  switch (strid(strlower(from))) {
    case strid("regular"):
      to = GFont::regular.value();
      break;
    case strid("bold"):
      to = GFont::bold.value();
      break;
    case strid("italic"):
      to = GFont::italic.value();
      break;
    case strid("big"):
    case strid("large"):
      to = GFont::large.value();
      break;
    case strid("small"):
      to = GFont::small.value();
      break;
    default:
      parseFont(from, to);
      break;
  }
  return true;
}

bool gconvert(GFontValue const& from, GString& to) {
  if (from.impl == GFontResource::emptyFont()) {
    to.clear();
    return false;
  }
  else {
    to = from.impl->style();
    to += (to.empty() ? "" : " ") + to_string(from.impl->size());
    to += (to.empty() ? "" : " ") + from.impl->families();
    return true;
  }
}

bool gconvert(GFontValue const& from, GFontValue& to) {
  to = from;
  return true;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GFont::GFont() = default;

GFont::GFont(GFontValue const& value) : Super(value) {}

GFont::GFont(GString const& value) {set(value);}

GFont::GFont(GString const& value, int) {  // for creating static vars
  pmodes_.isSet = true;
  gconvert(value, value_);
}

GFont* GFont::clone(bool copy_value) const {
  return copy_value ? new GFont(value_) : new GFont;
}

void GFont::updateGadget(Gadget& g) {
  g.updateLayout();
}

void GFont::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Font]) return;

  if (value_.size() > 0) {
    r.specif[GPropType::Font] = specif;
    r.font_ = value_;
  }
}

}
