//
// Font Prop
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitFont_hpp
#define GuitFont_hpp
#include <gvarprop.hpp>
namespace guit {

#if defined(_WIN32)
#  undef small
#endif

/// value of a GFont.
class GFontValue {
public:
  ~GFontValue();
  GFontValue();  
  GFontValue(GString const& fontspec);
  GFontValue(GFontValue const&);
  GFontValue(GFontValue const&&) = delete;

  GFontValue& operator=(GFontValue const&);

  bool operator==(GFontValue const&) const;
  bool operator!=(GFontValue const&) const;

  uint8_t size() const;
  float fontHeight() const;
  float fontAscent() const;
  float fontDescent() const;
  float fontWidth() const;   // size of letter C

  /// allocates resources, returns true if resources could be allocated.
  /// does not reallocate resources if already done.
  bool realize() const;

  /// returns the size of one line of text.
  /// does not take multiple lines into account, see GText for that purpose.
  GDim textSize(GString const& text, bool richtext, GTextAttributes const* = {}) const;

  mutable class GFontResource* impl{};
};


/// Font conversions
/// @{
bool gconvert(GFontValue const& from, GFontValue& to);
bool gconvert(GString const& from, GFontValue& to);
bool gconvert(GFontValue const& from, GString& to);
/// @}


/// Font Prop.
class GFont : public GVarProp_<GFontValue> {
public:
  using Super = GVarProp_<GFontValue>;

  /// Predefined Fonts.
  static GFont regular, bold, italic, large, small;

  GFont();

  /// Creates a Font Prop.
  /// Examples:
  /// - "12px Arial"
  /// - "Bold 12px Arial"
  /// - "Italic 12px Arial"
  GFont(GString const& fontname);
  
  GFont(GFontValue const& value);

  GFont* clone(bool copy_value) const override;

  /// Future value in an Active Expression.
  /// An Active Expression (GExpr) can refer ~prop.
  GPropRef operator~() {return GPropRef(*this);}

  GFont& operator=(GFont const& font) {set(font); return *this;}
  GFont& operator=(GString const& font) {set(font); return *this;}

  uint8_t size() const {return value_.size();}
  float fontHeight()  const {return value_.fontHeight();}
  float fontAscent()  const {return value_.fontAscent();}
  float fontDescent() const {return value_.fontDescent();}
  float fontWidth()   const {return value_.fontWidth();}   // size of letter C

  /// returns the size of one line of text.
  /// does not take multiple lines into account, see GText for that purpose.
  GDim textSize(GString const& str, bool richtext, GTextAttributes const* attr = {}) const {
    return value_.textSize(str, richtext, attr);
  }

  using GType = GPropType_<GFont,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  friend class Graphics;
  void apply(Gadget*, GRender&, Specif) final;
  void updateGadget(Gadget&) override;
private:
  GFont(GString const& fontname, int);  // for creating static vars
};

/// Returns a GFont created in the heap.
/// @{
inline GFont& Font() {return *new GFont();}
inline GFont& Font(GString const& value) {return *new GFont(value);}
/// @}

}
#endif
