//
// Text, RegText and Tip Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Text_hpp
#define Guit_Text_hpp
#include <gvarprop.hpp>
namespace guit {

/// Text prop.
class GText : public GVarProp_<GString> {
public:
  using Super = GVarProp_<GString>;

  GText();
  GText(GString const& value);
  GText(int value);
  GText(float value);
  GText(GText const&);
  GText* clone(bool copy_value) const override;

  GText*  toText() override {return this;}
  bool    isString() const override {return true;}
  
  GString stringValue() const override {return value_;}
  float   floatValue() const override;

  /// Future value in an Active Expression (see GExpr).
  auto operator~() {return GPropRef_<GText>(*this);}

  /// Binds this prop to an Active Expression.
  GText& operator<<=(GExpr&);

  GText& operator=(GString const& str) {set(str,false); return *this;}
  GText& operator=(GText const& text) {set(text(),false); return *this;}
  GText& operator=(GInt const& num);
  GText& operator=(GFloat const& num);

  GText& operator+=(GText const& text) {append(text.value_); return *this;}
  GText& operator+=(GString const& str) {append(str); return *this;}
  GText& operator+=(GInt const& num);
  GText& operator+=(GFloat const& num);

  using Super::set;
  void set(GString const& str) override {set(str, false);}
  void set(GProp const&) override;

  /// returns the char at this position.
  /// checks bounds: returns 0 if _pos_ is out of bounds.
  char operator[](size_t pos) const;

  bool setByteAt(int bytepos, char ch);  // tenir compte d'unicode !!!!

  bool empty() const;
  size_t size() const;

  int compare(GString const& str) const noexcept;
  int compare(GText const& txt) const noexcept {return compare(txt.value_);}

  size_t find(GString const& str, size_t pos = 0) const noexcept;
  size_t rfind(GString const& str, size_t pos = GString::npos) const noexcept;

  bool append(GString const& str);
  bool append(GText const& txt) {return append(txt.value_);}

  bool insert(size_t pos, GString const& str);
  bool insert(size_t pos, GString const& str, size_t subpos, size_t sublen);
  bool insert(size_t pos, GText const& t) {return insert(pos, t.value_);}

  bool erase(size_t pos = 0, size_t len = GString::npos);
  GString substr(size_t pos = 0, size_t len = GString::npos) const;

  using GVarProp::read;
  void read(std::istream&) override;

  /// returns the size of the text.
  GDim textSize(GFont&, bool richtext, bool multilines,
                GTextAttributes const* = nullptr) const;

  /// converts position of character (UTF8 code point) to index (byte position) and vice-versa.
  /// - each UTF8 character  can occupy 1 to 4 bytes.
  /// - returns position of last byte/character if first argument is < 0.
  /// @{
  ssize_t posToIndex(ssize_t charpos, bool richtext) const;
  ssize_t indexToPos(ssize_t bytepos, bool richtext) const;
  /// @}

  /// returns index (byte position) of next or previous character.
  /// - each UTF8 character  can occupy 1 to 4 bytes.
  /// @{
  ssize_t nextIndex(ssize_t bytepos, bool richtext) const;
  ssize_t previousIndex(ssize_t bytepos, bool richtext) const;
  /// @}

  using GType = GPropType_<GText,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

  bool doChange(GString const& v) override;

protected:
  void apply(Gadget*, GRender&, Specif) override;
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void updateGadget(Gadget&) override;
  virtual bool edit(GString const& newval);
};

GString operator+(GString const& str, GText const& text);
GString operator+(GText const& text, GString const& str);
GString operator+(GText const& text, GInt& num);
GString operator+(GText const& text, GFloat& num);

/// Return a GText created in the heap.
/// @{
GText& Text();
GText& Text(GString const& value);
/// @}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Constrained Text.
/// Entered text is constrained by a Regex.
class GRegText : public GText {
public:
  using Super = GText;

  GRegText(GString const& value = "");
  GRegText(GRegText const&);
  GRegText* clone(bool copy_value) const override;

  using GText::operator<<=;
  GRegText& operator<<=(GExpr&);

  GRegText& operator=(GString const& val);
  GRegText& operator=(const GText&);
  
  void setFormat(GString const& regex);
  void setTimeFormat();
  void setDateFormat();
  void setNameFormat();
  void setPhoneFormat();

  using GType = GPropType_<GRegText,GText>;
  static GType& Type();
  GType& type() const override {return Type();}

  bool isValid(GString const& v) const override;

protected:
  GString pattern_;
};

/// Return a GRegText created in the heap.
/// @{
GRegText& RegText();
GRegText& RegText(GString const& value);
/// @}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Tool Tip.
class GTip : public GText {
public:
  GTip(GString const& value = "") : GText(value) {}
  GTip* clone(bool copy_value) const override;

  using GType = GPropType_<GTip,GText>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  void apply(Gadget*, GRender&, Specif) override;
  void updateGadget(Gadget&) override {}
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void addModes(GadgetModes&) override;
  void removeModes(GadgetModes&) override;
};

/// Return a GTip created in the heap.
/// @{
GTip& Tip();
GTip& Tip(GString const& value);
/// @}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/// Auxiliary textual data.
class GTextData : public GText {
public:
  GTextData(GString const& value = "") : GText(value) {}
  GTextData* clone(bool copy_value) const override;

  using GType = GPropType_<GTextData,GText>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  void apply(Gadget*, GRender&, Specif) override {}
  bool onAdd(Gadget*) override {return true;}
  void onRemove(Gadget*) override {}
};

/// Return a GTextData created in the heap.
/// @{
GTextData& TextData();
GTextData& TextData(GString const& value);
/// @}

}
#endif
